# -*- coding: utf-8 -*-
"""
@file
@brief Helpers, inspired from `utils.py <https://github.com/winpython/winpython/blob/master/winpython/utils.py>`_
"""
from __future__ import print_function

import os
import os.path as osp
import subprocess
import re
import sys
import locale


# =============================================================================
# Patch chebang line (courtesy of Christoph Gohlke)
# =============================================================================
def patch_shebang_line(fname, pad=b' ', fLOG=print):
    """
    Remove absolute path to python.exe in shebang lines.

    @param      python      python extractor
    @param      pad         pad
    @param      fLOG        logging function
    @return                 boolean, True if patched, False otherwise
    """
    if sys.version_info[0] == 2:
        shebang_line = re.compile(r"(#!.+pythonw?\\.exe)")  # Python2.7
    else:
        shebang_line = re.compile(b"(#!.+pythonw?\\.exe)")  # Python3+

    with open(fname, 'rb') as fh:
        content = fh.read()

    content = shebang_line.split(content, maxsplit=1)
    if len(content) != 3:
        return
    exe = os.path.basename(content[1][2:])
    content[1] = b'#!' + exe + (pad * (len(content[1]) - len(exe) - 2))
    content = b''.join(content)

    try:
        with open(fname, 'wb') as fh:
            fh.write(content)
            fLOG("[pymy] patched", fname)
        return True
    except Exception:
        fLOG("[pymy] failed to patch", fname)
        return False


def get_env(name, current=True):
    """
    Return HKCU/HKLM environment variable name and value

    @param      name        name to look for
    @param      current     switch between *HKEY_CURRENT_USER* (True) and *HKEY_LOCAL_MACHINE* (False)
    @return                 tuple (see below)

    For example, get_user_env('PATH') may returns::

        ('Path', u'C:\\Program Files\\Intel\\WiFi\\bin\\')
    """
    import winreg
    root = winreg.HKEY_CURRENT_USER if current else winreg.HKEY_LOCAL_MACHINE
    key = winreg.OpenKey(root, "Environment")
    for index in range(0, winreg.QueryInfoKey(key)[1]):
        try:
            value = winreg.EnumValue(key, index)
            if value[0].lower() == name.lower():
                # Return both value[0] and value[1] because value[0] could be
                # different from name (lowercase/uppercase)
                return value[0], value[1]
        except Exception:
            break


def set_env(name, value, current=True):
    """
    Set HKCU/HKLM environment variables


    @param      name        name to look for
    @param      current     switch between *HKEY_CURRENT_USER* (True) and *HKEY_LOCAL_MACHINE* (False)
    """
    import winreg
    root = winreg.HKEY_CURRENT_USER if current else winreg.HKEY_LOCAL_MACHINE
    key = winreg.OpenKey(root, "Environment")
    try:
        _x, key_type = winreg.QueryValueEx(key, name)
    except WindowsError:
        key_type = winreg.REG_EXPAND_SZ
    key = winreg.OpenKey(root, "Environment", 0, winreg.KEY_SET_VALUE)
    winreg.SetValueEx(key, name, 0, key_type, value)
    from win32gui import SendMessageTimeout
    from win32con import (HWND_BROADCAST, WM_SETTINGCHANGE,
                          SMTO_ABORTIFHUNG)
    SendMessageTimeout(HWND_BROADCAST, WM_SETTINGCHANGE, 0,
                       "Environment", SMTO_ABORTIFHUNG, 5000)


def create_shortcut(path, description, filename,
                    arguments="", workdir="", iconpath="", iconindex=0):
    """
    Create Windows shortcut (.lnk file)

    @param      path            where to store the link
    @param      description     description
    @param      filename        link name
    @param      arguments       arguments to store
    @param      workdir         working directory
    @para       iconpath        icon
    @param      iconindex       icon index
    @return                     filename
    """
    import pythoncom
    from win32com.shell import shell
    ilink = pythoncom.CoCreateInstance(shell.CLSID_ShellLink, None,
                                       pythoncom.CLSCTX_INPROC_SERVER,
                                       shell.IID_IShellLink)
    ilink.SetPath(path)
    ilink.SetDescription(description)
    if arguments:
        ilink.SetArguments(arguments)
    if workdir:
        ilink.SetWorkingDirectory(workdir)
    if iconpath or iconindex:
        ilink.SetIconLocation(iconpath, iconindex)
    # now save it.
    ipf = ilink.QueryInterface(pythoncom.IID_IPersistFile)
    if not filename.endswith('.lnk'):
        filename += '.lnk'
    filename = os.path.join(path, filename)
    ipf.Save(filename, 0)
    return filename


def decode_fs_string(string):
    """Convert string from file system charset to unicode"""
    charset = sys.getfilesystemencoding()
    if charset is None:
        charset = locale.getpreferredencoding()
    return string.decode(charset)


def exec_shell_cmd(args, path):
    """Execute shell command (*args* is a list of arguments) in *path*"""
    # print " ".join(args)
    process = subprocess.Popen(args, stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE, cwd=path, shell=True)
    return decode_fs_string(process.stdout.read())


def get_gcc_version(path):
    """Return version of the GCC compiler installed in *path*"""
    return exec_shell_cmd('gcc --version', path).splitlines()[0].split()[-1]


def get_r_version(path):
    """Return version of the R installed in *path*"""
    return exec_shell_cmd('dir ..\\README.R*', path).splitlines()[-3].split("-")[-1]


def get_julia_version(path):
    """Return version of the Julia installed in *path*"""
    return exec_shell_cmd('julia.exe -v', path).splitlines()[0].split(" ")[-1]


def python_query(cmd, path):
    """Execute Python command using the Python interpreter located in *path*"""
    res = exec_shell_cmd('python -c "%s"' % cmd, path).splitlines()
    if not res:
        raise Exception(
            "CMD:\n{0}\nRES:\n{1}\nPATH:\n{2}".format(cmd, res, path))
    return res[0]


def get_python_infos(path):
    """Return (version, architecture) for the Python distribution located in
    *path*. The version number is limited to MAJOR.MINOR, the architecture is
    an integer: 32 or 64"""
    is_64 = python_query('import sys; print(sys.maxsize > 2**32)', path)
    arch = {'True': 64, 'False': 32}.get(is_64, None)
    ver = python_query("import sys; print('%d.%d' % (sys.version_info.major, "
                       "sys.version_info.minor))", path)
    if re.match(r'([0-9]*)\.([0-9]*)', ver) is None:
        ver = None
    return ver, arch


def get_python_long_version(path):
    """Return long version (X.Y.Z) for the Python distribution located in
    *path*"""
    ver = python_query("import sys; print('%d.%d.%d' % "
                       "(sys.version_info.major, sys.version_info.minor,"
                       "sys.version_info.micro))", path)
    if re.match(r'([0-9]*)\.([0-9]*)\.([0-9]*)', ver) is None:
        ver = None
    return ver


# =============================================================================
# Patch sourcefile (instead of forking packages)
# =============================================================================
def patch_sourcefile(fname, in_text, out_text, silent_mode=False):
    """Replace a string in a source file"""
    import io
    if osp.isfile(fname) and not in_text == out_text:
        with io.open(fname, 'r') as fh:
            content = fh.read()
        new_content = content.replace(in_text, out_text)
        if not new_content == content:
            if not silent_mode:
                print("patching ", fname, "from", in_text, "to", out_text)
            with io.open(fname, 'wt') as fh:
                fh.write(new_content)

# =============================================================================
# Patch sourcelines (instead of forking packages)
# =============================================================================


def patch_sourcelines(fname, in_line_start, out_line, endline='\n', silent_mode=False):
    """Replace the middle of lines between in_line_start and endline """
    import io
    import os.path as osp
    if osp.isfile(fname):
        with io.open(fname, 'r') as fh:
            contents = fh.readlines()
            content = "".join(contents)
            for l in range(len(contents)):
                if contents[l].startswith(in_line_start):
                    begining, middle = in_line_start, contents[
                        l][len(in_line_start):]
                    ending = ""
                    if middle.find(endline) > 0:
                        ending = endline + \
                            endline.join(middle.split(endline)[1:])
                        middle = middle.split(endline)[0]
                    middle = out_line
                    new_line = begining + middle + ending
                    if not new_line == contents[l]:
                        if not silent_mode:
                            print(
                                "patching ", fname, " from\n", contents[l], "\nto\n", new_line)
                    contents[l] = new_line
            new_content = "".join(contents)
        if not new_content == content:
            # if not silent_mode:
            #    print("patching ", fname, "from", content, "to", new_content)
            with io.open(fname, 'wt') as fh:
                try:
                    fh.write(new_content)
                except Exception as e:
                    print("impossible to patch", fname, "from", content,
                          "to", new_content, " --- ", str(e).replace("\n", "--"))


WININST_PATTERN = (r'([a-zA-Z0-9\-\_]*|[a-zA-Z\-\_\.]*)-([0-9\.\-]*[a-z]*[0-9]?)(-Qt-([0-9\.]+))?.(win32|win\-amd64)' +
                   r'(-py([0-9\.]+))?(-setup)?\.exe')

# SOURCE_PATTERN defines what an acceptable source package name is
# As of 2014-09-08 :
#    - the wheel package format is accepte in source directory
#    - the tricky regexp is tuned also to support the odd jolib naming :
#         . joblib-0.8.3_r1-py2.py3-none-any.whl,
#         . joblib-0.8.3-r1.tar.gz

SOURCE_PATTERN = r'([a-zA-Z0-9\-\_\.]*)-([0-9\.\_]*[a-z]*[0-9]?)(\.zip|\.tar\.gz|\-(py[2-7]*|py[2-7]*\.py[2-7]*)\-none\-any\.whl)'

# WHEELBIN_PATTERN defines what an acceptable binary wheel package is
# "cp([0-9]*)" to replace per cp(34) for python3.4
# "win32|win\_amd64" to replace per "win\_amd64" for 64bit
WHEELBIN_PATTERN = r'([a-zA-Z0-9\-\_\.]*)-([0-9\.\_]*[a-z0-9\+]*[0-9]?)-cp([0-9]*)\-none\-(win32|win\_amd64)\.whl'


def get_source_package_infos(fname):
    """Return a tuple (name, version) of the Python source package"""
    match = re.match(SOURCE_PATTERN, osp.basename(fname))
    if match is not None:
        return match.groups()[:2]


def do_script(this_script, python_exe=None,
              verbose=False, install_options=None):
    """Execute a script (get-pip typically)"""
    if python_exe is None:
        python_exe = sys.executable
    assert osp.isfile(python_exe)
    myroot = os.path.dirname(python_exe)

    # cmd = [python_exe, myroot + r'\Scripts\pip-script.py', 'install']
    cmd = [python_exe]
    if install_options:
        cmd += install_options  # typically ['--no-deps']
        print('script install_options', install_options)
    cmd += [this_script]
    # print('build_wheel', myroot, cmd)
    print("Executing ", cmd)

    if verbose:
        subprocess.call(cmd, cwd=myroot)
    else:
        p = subprocess.Popen(cmd, cwd=myroot, stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        p.communicate()
        p.stdout.close()
        p.stderr.close()
    if verbose:
        print("Executed " % cmd)
    return 'ok'


KEY_C = r"Software\Classes\%s"
KEY_C0 = KEY_C % r"Python.%sFile\shell"
KEY_C1 = KEY_C % r"Python.%sFile\shell\%s"
KEY_C2 = KEY_C1 + r"\command"
KEY_DROP0 = KEY_C % r"Python.%sFile\shellex"
KEY_DROP1 = KEY_C % r"Python.%sFile\shellex\DropHandler"
KEY_I = KEY_C % r"Python.%sFile\DefaultIcon"
KEY_D = KEY_C % r"Python.%sFile"
EWI = "Edit with IDLE"
EWS = "Edit with Spyder"

KEY_S = r"Software\Python"
KEY_S0 = KEY_S + r"\PythonCore"
KEY_S1 = KEY_S0 + r"\%s"


def register(target, current=True):
    """Register a Python distribution in Windows registry"""
    import winreg
    root = winreg.HKEY_CURRENT_USER if current else winreg.HKEY_LOCAL_MACHINE

    # Extensions
    winreg.SetValueEx(winreg.CreateKey(root, KEY_C % ".py"),
                      "", 0, winreg.REG_SZ, "Python.File")
    winreg.SetValueEx(winreg.CreateKey(root, KEY_C % ".pyw"),
                      "", 0, winreg.REG_SZ, "Python.NoConFile")
    winreg.SetValueEx(winreg.CreateKey(root, KEY_C % ".pyc"),
                      "", 0, winreg.REG_SZ, "Python.CompiledFile")
    winreg.SetValueEx(winreg.CreateKey(root, KEY_C % ".pyo"),
                      "", 0, winreg.REG_SZ, "Python.CompiledFile")

    # MIME types
    winreg.SetValueEx(winreg.CreateKey(root, KEY_C % ".py"),
                      "Content Type", 0, winreg.REG_SZ, "text/plain")
    winreg.SetValueEx(winreg.CreateKey(root, KEY_C % ".pyw"),
                      "Content Type", 0, winreg.REG_SZ, "text/plain")

    # Verbs
    python = osp.abspath(osp.join(target, 'python.exe'))
    pythonw = osp.abspath(osp.join(target, 'pythonw.exe'))
    spyder = osp.abspath(osp.join(target, os.pardir, 'Spyder.exe'))
    if not osp.isfile(spyder):
        spyder = '%s" "%s\\Scripts\\spyder' % (pythonw, target)
    winreg.SetValueEx(winreg.CreateKey(root, KEY_C2 % ("", "open")),
                      "", 0, winreg.REG_SZ, '"%s" "%%1" %%*' % python)
    winreg.SetValueEx(winreg.CreateKey(root, KEY_C2 % ("NoCon", "open")),
                      "", 0, winreg.REG_SZ, '"%s" "%%1" %%*' % pythonw)
    winreg.SetValueEx(winreg.CreateKey(root, KEY_C2 % ("Compiled", "open")),
                      "", 0, winreg.REG_SZ, '"%s" "%%1" %%*' % python)
    winreg.SetValueEx(winreg.CreateKey(root, KEY_C2 % ("", EWI)),
                      "", 0, winreg.REG_SZ,
                      '"%s" "%s\\Lib\\idlelib\\idle.pyw" -n -e "%%1"'
                      % (pythonw, target))
    winreg.SetValueEx(winreg.CreateKey(root, KEY_C2 % ("NoCon", EWI)),
                      "", 0, winreg.REG_SZ,
                      '"%s" "%s\\Lib\\idlelib\\idle.pyw" -n -e "%%1"'
                      % (pythonw, target))
    winreg.SetValueEx(winreg.CreateKey(root, KEY_C2 % ("", EWS)),
                      "", 0, winreg.REG_SZ, '"%s" "%%1"' % spyder)
    winreg.SetValueEx(winreg.CreateKey(root, KEY_C2 % ("NoCon", EWS)),
                      "", 0, winreg.REG_SZ, '"%s" "%%1"' % spyder)

    # Drop support
    handler = "{60254CA5-953B-11CF-8C96-00AA00B8708C}"
    for ftype in ("", "NoCon", "Compiled"):
        winreg.SetValueEx(winreg.CreateKey(root, KEY_DROP1 % ftype),
                          "", 0, winreg.REG_SZ, handler)

    # Icons
    dlls = osp.join(target, 'DLLs')
    winreg.SetValueEx(winreg.CreateKey(root, KEY_I % ""),
                      "", 0, winreg.REG_SZ, r'%s\py.ico' % dlls)
    winreg.SetValueEx(winreg.CreateKey(root, KEY_I % "NoCon"),
                      "", 0, winreg.REG_SZ, r'%s\py.ico' % dlls)
    winreg.SetValueEx(winreg.CreateKey(root, KEY_I % "Compiled"),
                      "", 0, winreg.REG_SZ, r'%s\pyc.ico' % dlls)

    # Descriptions
    winreg.SetValueEx(winreg.CreateKey(root, KEY_D % ""),
                      "", 0, winreg.REG_SZ, "Python File")
    winreg.SetValueEx(winreg.CreateKey(root, KEY_D % "NoCon"),
                      "", 0, winreg.REG_SZ, "Python File (no console)")
    winreg.SetValueEx(winreg.CreateKey(root, KEY_D % "Compiled"),
                      "", 0, winreg.REG_SZ, "Compiled Python File")

    # PythonCore entries
    '''
    short_version = utils.get_python_infos(target)[0]
    long_version = utils.get_python_long_version(target)
    key_core = (KEY_S1 % short_version) + r'\\%s'
    winreg.SetValueEx(winreg.CreateKey(root, key_core % 'InstallPath'),
                      "", 0, winreg.REG_SZ, target)
    winreg.SetValueEx(winreg.CreateKey(root,
                                       key_core % r'InstallPath\\InstallGroup'),
                      "", 0, winreg.REG_SZ, "Python %s" % short_version)
    winreg.SetValueEx(winreg.CreateKey(root, key_core % 'Modules'),
                      "", 0, winreg.REG_SZ, "")
    winreg.SetValueEx(winreg.CreateKey(root, key_core % 'PythonPath'),
                      "", 0, winreg.REG_SZ,
                      r"%s\\Lib;%s\\DLLs" % (target, target))
    winreg.SetValueEx(winreg.CreateKey(root,
                                       key_core % r'Help\\Main Python Documentation'),
                      "", 0, winreg.REG_SZ,
                      r"%s\\Doc\\python%s.chm" % (target, long_version))
    '''

    # Create start menu entries for all WinPython launchers
    '''
    for path, desc, fname in _get_shortcut_data(target, current=current):
        utils.create_shortcut(path, desc, fname)
    '''

    # Register the Python ActiveX Scripting client (requires pywin32)
    axscript = osp.join(target, 'Lib', 'site-packages', 'win32comext',
                        'axscript', 'client', 'pyscript.py')
    if osp.isfile(axscript):
        subprocess.call('"%s" "%s"' % (python, axscript), cwd=target)
    else:
        print('Unable to register ActiveX: please install pywin32',
              file=sys.stderr)


'''
def unregister(target, current=True):
    """Unregister a Python distribution in Windows registry"""
    # Registry entries
    root = winreg.HKEY_CURRENT_USER if current else winreg.HKEY_LOCAL_MACHINE
    short_version = utils.get_python_infos(target)[0]
    key_core = (KEY_S1 % short_version) + r'\\%s'
    for key in (
        # Drop support
        KEY_DROP1 % "", KEY_DROP1 % "NoCon", KEY_DROP1 % "Compiled",
        KEY_DROP0 % "", KEY_DROP0 % "NoCon", KEY_DROP0 % "Compiled",
        # Icons
        KEY_I % "NoCon", KEY_I % "Compiled", KEY_I % "",
        # Edit with IDLE
                KEY_C2 % ("", EWI), KEY_C2 % ("NoCon", EWI),
                KEY_C1 % ("", EWI), KEY_C1 % ("NoCon", EWI),
                # Edit with Spyder
                KEY_C2 % ("", EWS), KEY_C2 % ("NoCon", EWS),
                KEY_C1 % ("", EWS), KEY_C1 % ("NoCon", EWS),
                # Verbs
                KEY_C2 % ("", "open"),
                KEY_C2 % ("NoCon", "open"),
                KEY_C2 % ("Compiled", "open"),
                KEY_C1 % ("", "open"),
                KEY_C1 % ("NoCon", "open"),
                KEY_C1 % ("Compiled", "open"),
                KEY_C0 % "", KEY_C0 % "NoCon", KEY_C0 % "Compiled",
                # Descriptions
                KEY_D % "NoCon", KEY_D % "Compiled", KEY_D % "",
                # PythonCore
                key_core % r'InstallPath\\InstallGroup',
                key_core % 'InstallPath',
                key_core % 'Modules',
                key_core % 'PythonPath',
                key_core % r'Help\\Main Python Documentation',
                key_core % 'Help',
                KEY_S1 % short_version, KEY_S0, KEY_S,
    ):
        try:
            print(key)
            winreg.DeleteKey(root, key)
        except WindowsError:
            rootkey = 'HKEY_CURRENT_USER' if current else 'HKEY_LOCAL_MACHINE'
            print(r'Unable to remove %s\\%s' % (rootkey, key), file=sys.stderr)

    # Start menu shortcuts
    for path, desc, fname in _get_shortcut_data(target, current=current):
        if osp.exists(fname):
            os.remove(fname)
'''



# -*- coding: utf-8 -*-
"""
@brief      test log(time=700s)
"""
import os
import unittest
import pyquickhelper
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pycode import get_temp_folder, skipif_circleci
from pyquickhelper.ipythonhelper import execute_notebook_list, execute_notebook_list_finalize_ut
import pymyinstall


class TestRunNotebooks4(unittest.TestCase):

    def a_test_run_notebook(self, name):
        temp = get_temp_folder(
            __file__, "temp_run_notebooks4_{0}".format(name))

        fnb = os.path.normpath(os.path.join(
            os.path.abspath(os.path.dirname(__file__)), "..", "..", "_doc", "notebooks"))
        keepnote = []
        pyq = os.path.join(os.path.dirname(pyquickhelper.__file__), '..')
        code_init = "import sys\nsys.path.append('{0}')\n".format(
            pyq.replace("\\", "\\\\"))
        for f in os.listdir(fnb):
            if name not in f:
                continue
            if os.path.splitext(f)[-1] == ".ipynb":
                if "install_module" in f:
                    continue
                keepnote.append((os.path.join(fnb, f), code_init))
        self.assertTrue(len(keepnote) > 0)

        def valid(cell):
            if "snakeviz" in cell:
                return False
            return True

        addpaths = [os.path.normpath(os.path.join(
            os.path.abspath(os.path.dirname(__file__)), "..", "..", "src")),
            os.path.normpath(os.path.join(
                os.path.abspath(os.path.dirname(__file__)), "..", "..", "..", "jyquickhelper", "src"))
        ]

        if len(keepnote) == 0:
            raise ValueError("No notebook for name='{0}'".format(name))

        res = execute_notebook_list(
            temp, keepnote, fLOG=fLOG, valid=valid, additional_path=addpaths)
        execute_notebook_list_finalize_ut(
            res, fLOG=fLOG, dump=pymyinstall)

    @skipif_circleci('too long')
    def test_notebook_example_xgboost(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        self.a_test_run_notebook("example_xgboost")


if __name__ == "__main__":
    unittest.main()



"""
@file
@brief Creates batch file to set up the environment
"""
from __future__ import print_function

import os
import sys

if sys.version_info[0] == 2:
    from codecs import open


def create_win_batches(folders, verbose=False, selection=None, fLOG=print, module_list=None):
    """
    create batchs for the setup, they will be placed in
    *folders["config"]*

    @param      folders     dictionary with the keys *tools*, *config*, *python*, *workspace*
    @param      verbose     verbose
    @param      fLOG        logging function
    @param      selection   list of batchs to create
    @param      module_list list of python modules to install, to know which script to install or not
    @return                 operations (list of what was done)

    .. versionchanged:: 1.1
        Add batch file for glue and to run some checkings.
    """
    if selection is None:
        raise ValueError("selection cannot be None")
    if module_list is None:
        raise ValueError("module_list cannot be None")

    has_jupyter = False
    has_spyder = False
    has_rss = False
    has_glue = False
    has_orange = False
    has_sqlite_bro = False
    for mod in module_list:
        if mod.name == "jupyter":
            has_jupyter = True
        if mod.name == "orange3":
            has_orange = True
        if mod.name == "spyder":
            has_spyder = True
        if mod.name == "pyrsslocal":
            has_rss = True
        if mod.name == "glueviz":
            has_glue = True
        if mod.name == "sqlite_bro":
            has_sqlite_bro = True

    list_functions = [create_win_env,
                      create_win_scite,
                      create_win_putty,
                      create_win_sqllitespy,
                      create_win_python_console,
                      update_all_packages,
                      run_checkings,
                      win_replace_shebang,
                      win_check_installation,
                      ]

    if has_jupyter:
        list_functions.extend([create_win_jupyter_console,
                               create_win_jupyter_qtconsole,
                               create_win_jupyter_notebook,
                               win_install_kernels,
                               ])

    if has_orange:
        list_functions.append(create_win_orange)

    if has_spyder:
        list_functions.append(create_win_spyder)

    if has_rss:
        list_functions.append(create_win_rss)

    if has_sqlite_bro:
        list_functions.append(create_win_sqlite_bro)

    if has_glue:
        list_functions.append(create_win_glue)

    if "r" in selection:
        list_functions.append((create_win_r_console, "r"))
        list_functions.append((create_win_r_gui, "r"))

    if "julia" in selection:
        list_functions.append((create_win_julia_console, "julia"))

    operations = []
    for func in list_functions:
        if isinstance(func, tuple):
            func, name = func
        else:
            name = None

        if name is None or name in selection:
            op = func(folders)
            if verbose:
                for o in op:
                    fLOG("[pymy] " + " ".join(o))
            operations.extend(op)
    return operations


def create_win_env(folders):
    """
    create a batch file to set up the environment

    @param      folders     see @see fn create_win_batches
    @return                 operations (list of what was done)
    """
    tools = folders["tools"]
    text = ['@echo off',
            'set CURRENT=%~dp0',
            'set PYTHON_TOOLS=%CURRENT%..\\tools',
            'set PYTHON_WINHOME=%CURRENT%..\\python',
            'set PYTHON_WINSCRIPTS=%CURRENT%..\\python\\Scripts',
            'set WORKSPACE=%CURRENT%..\\workspace',
            'set PATH=%PYTHON_WINHOME%;%PYTHON_WINSCRIPTS%;%PATH%']
    if os.path.exists(os.path.join(tools, "R")):
        text.append('set R_HOME=%PYTHON_TOOLS%\\R')
        text.append('set R_LIBS=%PYTHON_TOOLS%\\R\\library')
    if os.path.exists(os.path.join(tools, "Julia")):
        text.append('set JULIA_HOME=%PYTHON_TOOLS%\\Julia')
        text.append('set JULIA_PKGDIR=%PYTHON_TOOLS%\\Julia\\pkg')
    if os.path.exists(os.path.join(tools, "TDM")):
        text.append('set PATH=%PATH%;%PYTHON_TOOLS%\\TDM\\bin')
    if os.path.exists(os.path.join(tools, "MinGW")):
        text.append('set PATH=%PATH%;%PYTHON_TOOLS%\\MinGW\\bin')
    if os.path.exists(os.path.join(tools, "Graphviz")):
        text.append('set PATH=%PATH%;%PYTHON_TOOLS%\\Graphviz\\bin')
    if os.path.exists(os.path.join(tools, "JDK")):
        text.append('set PATH=%PATH%;%PYTHON_TOOLS%\\JDK\\bin')

    text = "\n".join(text)
    name = os.path.join(folders["config"], "env.bat")
    with open(name, "w") as f:
        f.write(text)
    return [('batch', name)]


def create_win_jupyter_console(folders):
    """
    create a batch file to start jupyter

    @param      folders     see @see fn create_win_batches
    @return                 operations (list of what was done)
    """
    text = ['@echo off',
            'set CURRENT2=%~dp0',
            'call "%CURRENT2%env.bat"',
            'set JUPYTERC=%PYTHON_WINSCRIPTS%\\jupyter-console.exe',
            '"%JUPYTERC%" console']
    # command jupyter console does not work yet even if the documentation says
    # so

    text = "\n".join(text)
    name = os.path.join(folders["config"], "jupyter_console.bat")
    with open(name, "w") as f:
        f.write(text)
    return [('batch', name)]


def create_win_orange(folders):
    """
    create a batch file to start orange

    @param      folders     see @see fn create_win_batches
    @return                 operations (list of what was done)
    """
    text = ['@echo off',
            'set CURRENT2=%~dp0',
            'call "%CURRENT2%env.bat"',
            'set ORANGE=%PYTHON_WINSCRIPTS%\\orange-canvas.exe',
            '"%ORANGE%"']
    # command jupyter console does not work yet even if the documentation says
    # so

    text = "\n".join(text)
    name = os.path.join(folders["config"], "orange_canvas.bat")
    with open(name, "w") as f:
        f.write(text)
    return [('batch', name)]


def create_win_jupyter_qtconsole(folders):
    """
    create a batch file to start Jupyter QtConsole

    @param      folders     see @see fn create_win_batches
    @return                 operations (list of what was done)
    """
    text = ['@echo off',
            'set CURRENT2=%~dp0',
            'call "%CURRENT2%env.bat"',
            'set JUPYTERQTC=%PYTHON_WINSCRIPTS%\\jupyter-qtconsole.exe',
            'start "jupyter-qtconsole" /B "%JUPYTERQTC%"']

    text = "\n".join(text)
    name = os.path.join(folders["config"], "jupyter_qtconsole.bat")
    with open(name, "w") as f:
        f.write(text)
    return [('batch', name)]


def create_win_jupyter_notebook(folders):
    """
    create a batch file to start Jupyter Notebook

    @param      folders     see @see fn create_win_batches
    @return                 operations (list of what was done)
    """
    text = ['@echo off',
            'set CURRENT2=%~dp0',
            'call "%CURRENT2%env.bat"',
            'set JUPYTERNB=%PYTHON_WINSCRIPTS%\\jupyter-notebook.exe',
            '"%JUPYTERNB%" "--notebook-dir=%WORKSPACE%" "--config=%CURRENT2%profile_win_profile\\ipython_kernel_config.py"']

    text = "\n".join(text)
    name = os.path.join(folders["config"], "jupyter_notebook.bat")
    with open(name, "w") as f:
        f.write(text)
    return [('batch', name)]


def create_win_scite(folders):
    """
    create a batch file to start scite

    @param      folders     see @see fn create_win_batches
    @return                 operations (list of what was done)
    """
    text = ['@echo off',
            'set CURRENT2=%~dp0',
            'call "%CURRENT2%env.bat"',
            'set SCITE=%PYTHON_TOOLS%\\Scite\\wscite\\scite.exe',
            'start "scite" /B "%SCITE%" "%1"']

    text = "\n".join(text)
    name = os.path.join(folders["config"], "scite.bat")
    with open(name, "w") as f:
        f.write(text)
    return [('batch', name)]


def create_win_putty(folders):
    """
    create a batch file to start scite

    @param      folders     see @see fn create_win_batches
    @return                 operations (list of what was done)
    """
    text = ['@echo off',
            'set CURRENT2=%~dp0',
            'call %CURRENT2%env.bat',
            'set PUTTY=%PYTHON_TOOLS%\\Putty\\putty.exe',
            '"start "putty" /B "%PUTTY%"']

    text = "\n".join(text)
    name = os.path.join(folders["config"], "putty.bat")
    with open(name, "w") as f:
        f.write(text)
    return [('batch', name)]


def create_win_sqllitespy(folders):
    """
    create a batch file to start sqlitespy

    @param      folders     see @see fn create_win_batches
    @return                 operations (list of what was done)
    """
    text = ['@echo off',
            'set CURRENT2=%~dp0',
            'call "%CURRENT2%env.bat"',
            'set SQLITESPY=%PYTHON_TOOLS%\\SQLiteSpy\\SQLiteSpy.exe',
            'cd "%WORKSPACE%"',
            'start "SqliteSpy" /B "%SQLITESPY%"']

    text = "\n".join(text)
    name = os.path.join(folders["config"], "sqlitespy.bat")
    with open(name, "w") as f:
        f.write(text)
    return [('batch', name)]


def create_win_python_console(folders):
    """
    create a batch file to start python

    @param      folders     see @see fn create_win_batches
    @return                 operations (list of what was done)
    """
    text = ['@echo off',
            'set CURRENT2=%~dp0',
            'call "%CURRENT2%env.bat"',
            'set PYTHON=%PYTHON_WINHOME%\\python.exe',
            '"%PYTHON%"']

    text = "\n".join(text)
    name = os.path.join(folders["config"], "python_console.bat")
    with open(name, "w") as f:
        f.write(text)
    return [('batch', name)]


def create_win_julia_console(folders):
    """
    create a batch file to start julia

    @param      folders     see @see fn create_win_batches
    @return                 operations (list of what was done)
    """
    text = ['@echo off',
            'set CURRENT2=%~dp0',
            'call "%CURRENT2%env.bat"',
            'set JULIA=%PYTHON_TOOLS%\\Julia\\bin\\julia.exe',
            '"%JULIA%"']

    text = "\n".join(text)
    name = os.path.join(folders["config"], "julia_console.bat")
    with open(name, "w") as f:
        f.write(text)
    return [('batch', name)]


def create_win_spyder(folders):
    """
    create a batch file to start spyder

    @param      folders     see @see fn create_win_batches
    @return                 operations (list of what was done)

    .. index:: Spyder, PySide, PyQt, PySide2

    This installation uses `PySide2 <https://pypi.python.org/pypi/PySide2/>`_
    instead of `PyQt <https://www.riverbankcomputing.com/software/pyqt/intro>`_.

        set QT_API=pyside
    """
    text = ['@echo off',
            'set CURRENT2=%~dp0',
            'call "%CURRENT2%env.bat"',
            'set QT_API=pyside',
            '"%PYTHON_WINSCRIPTS%\\spyder.bat" "--workdir=%WORKSPACE%"']

    text = "\n".join(text)
    name = os.path.join(folders["config"], "spyder.bat")
    with open(name, "w") as f:
        f.write(text)
    return [('batch', name)]


def create_win_r_console(folders):
    """
    create a batch file to start R

    @param      folders     see @see fn create_win_batches
    @return                 operations (list of what was done)
    """
    text = ['@echo off',
            'set CURRENT2=%~dp0',
            'call "%CURRENT2%env.bat"',
            'set REXE=%PYTHON_TOOLS%\\R\\bin\\x64\\R.exe',
            '"%REXE%"']

    text = "\n".join(text)
    name = os.path.join(folders["config"], "r_console.bat")
    with open(name, "w") as f:
        f.write(text)
    return [('batch', name)]


def create_win_r_gui(folders):
    """
    create a batch file to start R Gui

    @param      folders     see @see fn create_win_batches
    @return                 operations (list of what was done)
    """
    text = ['@echo off',
            'set CURRENT2=%~dp0',
            'call "%CURRENT2%env.bat"',
            'set RGUI=%PYTHON_TOOLS%\\R\\bin\\x64\\Rgui.exe',
            'start "RGui" /B "%RGUI%"']

    text = "\n".join(text)
    name = os.path.join(folders["config"], "r_gui.bat")
    with open(name, "w") as f:
        f.write(text)
    return [('batch', name)]


def win_install_kernels(folders, suffix=""):
    """
    create a batch file to install kernels

    @param      folders     see @see fn create_win_batches
    @param      suffix      add a suffix
    @return                 operations (list of what was done)
    """
    text = ['@echo off',
            'set CURRENT2=%~dp0',
            'call "%CURRENT2%env.bat"',
            ('"%PYTHON_WINHOME%\\pythonw.exe" -u -c "from pymyinstall.win_installer ' +
             'import inno_install_kernels;inno_install_kernels(\'CURRENT2\', \'%1\')"')]

    text = "\n".join(text)
    name = os.path.join(folders["config"], "add_kernels.bat")
    with open(name, "w") as f:
        f.write(text)
    return [('batch', name)]


def win_replace_shebang(folders, suffix=""):
    """
    create a batch file to replace the shebang

    @param      folders     see @see fn create_win_batches
    @param      suffix      add a suffix
    @return                 operations (list of what was done)
    """
    text = ['@echo off',
            'set CURRENT2=%~dp0',
            'call "%CURRENT2%env.bat"',
            '"%PYTHON_WINHOME%\\pythonw.exe" -u -c "import os;' +
            'from pymyinstall.win_installer import win_patch_paths;win_patch_paths(\'PYTHON_WINSCRIPTS\', None)"']

    text = "\n".join(text)
    name = os.path.join(folders["config"], "replace_shebang.bat")
    with open(name, "w") as f:
        f.write(text)
    return [('batch', name)]


def win_check_installation(folders, suffix=""):
    """
    create a batch file to check the installation when well

    @param      folders     see @see fn create_win_batches
    @param      suffix      add a suffix (unused)
    @return                 operations (list of what was done)
    """
    text = ['@echo off',
            'set CURRENT2=%~dp0',
            'call "%CURRENT2%env.bat"',
            '"%PYTHON_WINHOME%\\python.exe" -u -c "import sys;from pymyinstall.win_installer import import_every_module;' +
            'import_every_module(sys.executable, None, fLOG=print)"']

    text = "\n".join(text)
    name = os.path.join(folders["config"], "win_check_installation.bat")
    with open(name, "w") as f:
        f.write(text)
    return [('batch', name)]


def create_win_rss(folders, suffix=""):
    """
    create a batch file to start RSS reader

    @param      folders     see @see fn create_win_batches
    @param      suffix      add a suffix
    @return                 operations (list of what was done)
    """
    text = ['@echo off',
            'set CURRENT2=%~dp0',
            'call "%CURRENT2%env.bat"',
            '"%PYTHON_WINHOME%\\python.exe" -u -c "from pyquickhelper.loghelper import fLOG;' +
            'from pyquickhelper.pycode.blog_helper import rss_update_run_server;fLOG(OutputPrint=True);' +
            'rss_update_run_server(r\'%CURRENT2%rss_database.db3\', r\'%CURRENT2%rss_list.xml\')"']

    text = "\n".join(text)
    name = os.path.join(folders["config"], "run_fetch_rss.bat")
    with open(name, "w") as f:
        f.write(text)

    text = """
                <?xml version="1.0" encoding="UTF-8"?>
                <opml version="1.0">
                    <head>
                        <title>teachings subscriptions</title>
                    </head>
                    <body>
                        <outline text="XD blog" title="XD blog" type="rss"
                            xmlUrl="http://www.xavierdupre.fr/blog/xdbrss.xml"
                            htmlUrl="http://www.xavierdupre.fr/blog/xd_blog.html" />
                    </body>
                </opml>
                """.replace('                ', "")

    rss_name = os.path.join(folders["config"], "rss_list.xml")
    with open(rss_name, 'w', encoding="utf8") as f:
        f.write(text)

    return [('batch', name), ('rss', rss_name)]


def create_win_glue(folders, suffix=""):
    """
    create a batch file to start Glue

    @param      folders     see @see fn create_win_batches
    @param      suffix      add a suffix
    @return                 operations (list of what was done)

    .. versionadded:: 1.1
    """
    text = ['@echo off',
            'set CURRENT2=%~dp0',
            'call "%CURRENT2%env.bat"',
            '"%PYTHON_WINSCRIPTS%\\glue.exe" %1']

    text = "\n".join(text)
    name = os.path.join(folders["config"], "run_glue.bat")
    with open(name, "w") as f:
        f.write(text)
    return [('batch', name)]


def create_win_sqlite_bro(folders, suffix=""):
    """
    create a batch file to start SqliteBro

    @param      folders     see @see fn create_win_batches
    @param      suffix      add a suffix
    @return                 operations (list of what was done)

    .. versionadded:: 1.1
    """
    text = ['@echo off',
            'set CURRENT2=%~dp0',
            'call "%CURRENT2%env.bat"',
            '"%PYTHON_WINSCRIPTS%\\sqlite_bro.exe" %1']

    text = "\n".join(text)
    name = os.path.join(folders["config"], "sqlite_bro.bat")
    with open(name, "w") as f:
        f.write(text)
    return [('batch', name)]


def update_all_packages(folders, suffix=""):
    """
    create a batch file to update all packages

    @param      folders     see @see fn create_win_batches
    @param      suffix      add a suffix
    @return                 operations (list of what was done)

    .. versionchanged:: 1.1
         Bug fix, update script to import function update_all (fails in 1.2).
    """
    text = ['@echo off',
            'set CURRENT2=%~dp0',
            'call "%CURRENT2%env.bat"',
            '"%PYTHON_WINHOME%\\python.exe" -u -c "from pymyinstall.packaged import update_all;' +
            'update_all(temp_folder=\'%WORKSPACE%/update_modules\', verbose=True)"']

    text = "\n".join(text)
    name = os.path.join(folders["config"], "run_update_all_packages.bat")
    with open(name, "w") as f:
        f.write(text)
    return [('batch', name)]


def run_checkings(folders, suffix=""):
    """
    create a batch file to update all packages

    @param      folders     see @see fn create_win_batches
    @param      suffix      add a suffix
    @return                 operations (list of what was done)

    .. versionadded:: 1.1
    """
    text = ['@echo off',
            'set CURRENT2=%~dp0',
            'call "%CURRENT2%env.bat"',
            ('"%PYTHON_WINHOME%\\python.exe" -u -c "from pymyinstall.win_installer import ' +
             'distribution_checkings;distribution_checkings(None, None)"')]

    text = "\n".join(text)
    name = os.path.join(folders["config"], "run_checkings.bat")
    with open(name, "w") as f:
        f.write(text)
    return [('batch', name)]



# -*- coding: utf-8 -*-
"""
@file
@brief Defines a very small set of modules.
"""
import sys


def minimal_set():
    """
    list of modules to add to python to get a minimal python
    """
    names = [
        "autopep8",
        'entrypoints',
        "flake8",
        'markupsafe',
        "mccabe",
        "pep8",
        "pipdeptree",
        "psutil",
        "pyflakes",
        "pycodestyle",
        "pyquicksetup",
        "pywin32" if sys.platform.startswith("win") else None,
        "pywin32-ctypes" if sys.platform.startswith("win") else None,
        "six",
        "virtualenv",
        "wheel",
        "winshell" if sys.platform.startswith("win") else None,
    ]

    from .automate_install import find_module_install
    return [find_module_install(_) for _ in names if _ is not None]


def pywin32_set():
    """
    list of modules to add to python to get python with pywin32
    """
    names = ["pywin32" if sys.platform.startswith("win") else None,
             "pywin32-ctypes" if sys.platform.startswith("win") else None,
             "winshell" if sys.platform.startswith("win") else None,
             ]

    from .automate_install import find_module_install
    return [find_module_install(_) for _ in names if _ is not None]



"""
@file
@grief Functions to help using pywin32.
"""

from __future__ import print_function

import sys
import os
import shutil

if sys.version_info[0] == 2:
    FileNotFoundError = Exception


def import_pywin32():
    """
    For the module ``pywin32``,
    this function tries to add the path to the DLL to ``PATH``
    before throwing the exception:
    ``DLL load failed: The specified module could not be found``.

    It checks the instruction ``import win32com``.
    """
    try:
        import win32com
    except ImportError as e:
        if "DLL load failed:" in str(e):
            import os
            import sys
            import numpy
            from distutils.sysconfig import get_python_lib

            paths = set([os.path.join(
                os.path.split(sys.executable)[0], "Lib", "site-packages", "pywin32_system32"),
                os.path.join(get_python_lib(), "pywin32_system32"),
                os.path.join(
                    os.path.dirname(numpy.__file__), "..", "pywin32_system32"),
            ])

            epath = os.environ["PATH"]
            for path in paths:
                # exe = os.path.abspath(os.path.dirname(sys.executable))
                os.environ["PATH"] = epath + ";" + path

                try:
                    import win32com
                    return win32com
                except ImportError:
                    # we try the next path
                    pass

            try:
                import win32com
                return win32com
            except ImportError:
                # addition for WinPython
                exe = os.path.abspath(os.path.dirname(sys.executable))
                os.environ["PATH"] = os.environ["PATH"] + ";" + exe
                try:
                    import win32com
                    return win32com
                except ImportError:
                    dll = os.listdir(path)
                    dll = [os.path.join(path, _) for _ in dll if "dll" in _]
                    if len(dll) == 0:
                        raise ImportError("Did you install pywin32?") from e
                    else:
                        raise ImportError(
                            "Some DLL must be copied:\n" + "\n".join(dll)) from e
        else:
            raise e


def fix_pywin32_installation(python_path=None, fLOG=print):
    """
    copy DLL at the right place

    @param      python_path     python path
    @param      fLOG            logging function

    .. faqref::
        :title: pywin32 does not work

        To check module `pywin32 <https://pypi.python.org/pypi/pywin32>`_ is installed,
        you must run::

            import win32com

        If it displays the message ``ImportError: DLL load failed``, it means
        it was not able to find DLLs *pythoncom34.dll*, *pythoncom34.dll*.
        Two solutions:

        * Add the folder ``C:\\Python34_x64\\Lib\\site-packages\\pywin32_system32``
        to environment variable ``PATH``. That's what function
        @see fn import_pywin32 is doing every time it is called.
        * Copy the two DLLs to ``C:\\Windows\\System32``, that's what function
        @see fn fix_pywin32_installation does if it is run with admin rights.

    .. versionadded:: 1.1
    """
    if python_path is None:
        python_path = sys.executable.replace("w.exe", ".exe")
    if os.path.isfile(python_path):
        python_path = os.path.dirname(python_path)
    fdll = os.path.join(python_path, "Lib",
                        "site-packages", "pywin32_system32")
    dest_fold = [python_path, os.path.join(python_path, "DLLs"),
                 "C:\\Windows\\System32"]
    for dll in os.listdir(fdll):
        full = os.path.join(fdll, dll)
        if os.path.isdir(full):
            continue
        for destf in dest_fold:
            dest = os.path.join(destf, dll)
            if not os.path.exists(dest):
                shutil.copy(full, destf)
                fLOG("[pymy] copy", full, "to", destf)
            else:
                fLOG("[pymy] already copied", dest)



"""
@file
@brief Function to start on Windows from scratch

.. versionadded:: 1.1
"""
from __future__ import print_function
import sys
import os
from ..win_installer.win_setup_main_helper import win_download, win_install


def windows_default_tools_list():
    """
    returns a list of tools to install

    @return     list of tools
    """
    return ["7z", "scite", "putty", "mingw", "SQLiteSpy", "r", "vs",
            "julia", "graphviz", "tdm", "pandoc",
            "jdk", "jenkins",
            "miktex", "inkscape",
            "git", "python"]


def windows_startup(destination, temp_folder, params=None, fLOG=print):
    """
    Installs many tools.
    It does not work with Python 2.7.

    @param      destination     destination
    @param      temp_folder     temporary folder (not cleaned), for downloading purpose
    @param      fLOG            logging function
    @param      params          dictionary
    @return                     operations (what was done)

    About *params*, it can contains:

    * *tools*: list of tools to install, see @see fn win_download, if None, use a default list defined
      in @see fn windows_default_tools_list

    .. exref::
        :title: Setup a machine

        The following code tries to download many tools and packages
        to prepare a machine for a datascientist on Windows::

            from pymyinstall.startup import windows_startup
            windows_startup(r"d:\\datascientist", r"d:\\temp\\datascientist", fLOG=print)

    .. versionadded:: 1.1
    """
    if sys.version_info[0] == 2:
        raise SystemError("It requires Python 3+")

    # creating folders if they don't exists
    if not os.path.exists(destination):
        os.makedirs(destination)
    if not os.path.exists(temp_folder):
        os.makedirs(temp_folder)
    if params is None:
        params = {}

    # selection
    tools = params.get("tools", None)
    if tools is None:
        tools = windows_default_tools_list()

    if isinstance(tools, list):
        tools = {k: None for k in tools}

    fLOG("[pymy] ------ download", tools)
    op = win_download(temp_folder, fLOG=fLOG, selection=tools, module_list=[])

    # copy de SQLiteSpy, putty, Scite
    names = ["Julia", "Scite", "7z", "TDM", "MinGW", "R", "pandoc",
             "SQLiteSpy", "Putty", "Graphviz", "Jdk", "Jenkins", "Git",
             "MikTex", "InkScape",
             "Python", ]

    fLOG("[pymy] ------ install", tools)
    folders = dict(tools=destination,
                   python=os.path.join(destination, "python"))
    op += win_install(folders, download_folder=temp_folder,
                      fLOG=fLOG, selection=tools, names=names)
    return op



"""
@brief      test log(time=2s)
"""
import unittest
import warnings
from pyquickhelper.loghelper import fLOG


class TestLONGScriptInstall(unittest.TestCase):

    def test_pypi(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        import xmlrpc.client as xmlrpc_client
        module_name = "version_information"
        url = 'https://pypi.org/pypi/pip/json'
        functions = []
        with xmlrpc_client.ServerProxy(url) as pypi:
            try:
                for f in pypi.system.listMethods():
                    fLOG(f)
                    sig = pypi.system.methodSignature(f)
                    fLOG("    ", sig)
                    h = pypi.system.methodHelp(f)
                    fLOG("    ", h)
                    functions.append(f)
                    if len(functions) > 1:
                        break
                available = pypi.package_releases(module_name, True)
                fLOG(available)
            except xmlrpc_client.ProtocolError as e:
                warnings.warn("PyPI protocal has changed {0}".format(e))
                functions = [None, None]
            assert len(functions) > 1


if __name__ == "__main__":
    unittest.main()



"""
@file
@brief Various functions to install `MinGW <http://www.mingw.org/>`_.
"""
from __future__ import print_function
import sys
import os
import re

from .install_custom import download_page, download_file
from ..installhelper.install_cmd_helper import unzip_files


def install_operadriver(dest_folder=".", fLOG=print, install=True, version=None):
    """
    Installs `operadriver <https://github.com/operasoftware/operachromiumdriver/releases>`_.

    @param      dest_folder     where to download the setup
    @param      fLOG            logging function
    @param      install         install (otherwise only download)
    @param      version         version to install (unused)
    @return                     zip file in a list or list of unzipped files

    This is required for `Selenium <https://selenium-python.readthedocs.io/>`_.
    """
    if version is None:
        content = download_page(
            "https://github.com/operasoftware/operachromiumdriver/releases")
        reg = re.compile(
            "/tag/v([.][0-9]+[.][0-9]+([.][0-9]+)?([.][0-9]+)?)")
        f = reg.findall(content)
        if not f:
            raise Exception(
                "unable to get the last version number for OperaDriver")
        version = f[0][0]
    if sys.platform.startswith("win"):
        url = "https://github.com/operasoftware/operachromiumdriver/releases/download/v{0}/operadriver_win64.zip".format(
            version)
    elif sys.platform.startswith("mac"):
        url = "https://github.com/operasoftware/operachromiumdriver/releases/download/v{0}/operadriver_mac64.zip".format(
            version)
    else:
        url = "https://github.com/operasoftware/operachromiumdriver/releases/download/v{0}/operadriver_linux64.zip".format(
            version)
    name = url.split("/")[-1]

    outfile = os.path.join(dest_folder, name)
    fLOG("[pymy] operadriver, download from ", url)
    download_file(url, outfile, fLOG=fLOG)

    if install:
        return unzip_files(outfile, whereTo=dest_folder, fLOG=fLOG)
    else:
        return [outfile]



"""
@brief      test log(time=200s)
"""
import unittest
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pycode import ExtTestCase
from pymyinstall.packaged import small_set


class TestDifference(ExtTestCase):

    def test_diff(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        name = set(_.name for _ in small_set())
        keep = []
        for mod in small_set():
            if mod.name not in name:
                keep.append(mod)
        self.assertGreater(len(keep), 0)

        for mod in keep:
            if mod.mname is None:
                fLOG(
                    "ModuleInstall('{0}', '{1}'),".format(mod.name, mod.kind))
            else:
                fLOG("ModuleInstall('{0}', '{1}', mname='{2}'),".format(
                    mod.name, mod.kind, mod.mname))

    def test_diff2(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")
        res = small_set()
        count = {}
        for mod in res:
            count[mod.name] = 1
        self.assertIn("coverage", count)


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=1s)
"""
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall.installhelper import get_module_dependencies, get_module_metadata
from pymyinstall.installhelper.install_cmd_helper import is_conda_distribution
from pymyinstall.installhelper import missing_dependencies
from pymyinstall.installhelper.pip_helper import get_installed_distributions


class TestModuleDependenciesBug(unittest.TestCase):

    def common_function(self, name, use_pip=None):
        res = get_module_dependencies(name, deep=True, use_pip=use_pip)
        for k, v in sorted(res.items()):
            self.assertTrue(isinstance(v, tuple))
            fLOG(k, "-->", v)
        if len(res) < 3:
            pkgs = get_installed_distributions(
                local_only=False, skip=[])
            req_map = dict((p.key, (p, p.requires())) for p in pkgs)
            mat = req_map.get(name, None)
            raise AssertionError("len(res)={0}\nres={1}\ndata={2}\nmat={3}".format(
                len(res), res, get_module_metadata(name), mat))

    def test_dependencies_luigi(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")
        try:
            import luigi
            mod = "luigi"
            self.assertTrue(luigi is not None)
        except ImportError:
            mod = "pandas"

        if not is_conda_distribution():
            dep = missing_dependencies(mod)
            self.assertEqual(dep, {})
            self.common_function(mod)
        else:
            # issue with Anaconda, not investigated
            return


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=140s)

skip this test for regular run
"""

import sys
import os
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall.installcustom import install_pandoc


class TestLONGPandoc (unittest.TestCase):

    def test_install_pandoc(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")
        fold = os.path.abspath(os.path.split(__file__)[0])

        if sys.platform.startswith("win"):
            temp = os.path.join(fold, "temp_pandoc")
            if not os.path.exists(temp):
                os.mkdir(temp)
            for _ in os.listdir(temp):
                if ".msi" in _:
                    os.remove(os.path.join(temp, _))
            r = install_pandoc(
                temp_folder=temp,
                fLOG=fLOG,
                install=False,
                force_download=True)
            self.assertTrue(os.path.exists(r))


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
@file
@brief Defines different a set of IoT related modules (Internet of Things)
"""
import sys

from ..installhelper.module_install import ModuleInstall


def iot_set():
    """
    list of sphinx themes and others helpers for sphinx, it requires the modules in set *small*
    """
    mod = [
        ModuleInstall(
            'phue', 'pip', purpose="A Philips Hue Python library", usage="IoT"),
        ModuleInstall('ledcontroller', 'pip',
                      purpose="Controller library for limitlessled/easybulb/milight Wi-Fi LEDs", usage="IoT"),
        ModuleInstall('enum-compat', 'pip',
                      purpose="enum/enum34 compatibility package") if sys.version_info[0] == 2 else None,
        ModuleInstall('netifaces', 'pip',
                      purpose="Portable network interface information."),
        ModuleInstall('ifaddr', 'pip', purpose="for zero conf"),
        ModuleInstall('zeroconf', 'pip',
                      purpose="Multicast DNS Service Discovery for Python, originally by Paul Scott-Murphy."),
        ModuleInstall('casttube', 'pip', purpose="for pychromecast"),
        ModuleInstall('pychromecast', 'pip',
                      purpose="Library for Python 2 and 3 to communicate with the Google Chromecast."),
        ModuleInstall('python-xlib', 'pip', mname="Xlib",
                      purpose="Python X Library"),
        ModuleInstall('pyHook', 'wheel',
                      purpose="Python wrapper for global input hooks in Windows. The package provides callbacks "
                      + "for mouse and keyboard events;"
                      + " events can be monitored and filtered."),
        ModuleInstall('pyuserinput', 'pip', mname="pykeyboard",
                      purpose="A simple, cross-platform module for mouse and keyboard control"),
        ModuleInstall('tellcore-py', 'pip', mname="tellcore", usage="IoT",
                      purpose="Python wrapper for Telldus' home automation library"),
        ModuleInstall('python-nmap', 'pip', mname="nmap",
                      purpose="This is a python class to use nmap and access scan results from python3"),
        ModuleInstall('python-magic', 'pip',
                      purpose="File type identification using libmagic"),
        ModuleInstall('websocket-client', 'pip', mname="websocket",
                      purpose="WebSocket client for python. hybi13 is supported."),
        ModuleInstall('pushbullet.py', 'pip', usage="IoT",
                      purpose="A simple python client for pushbullet.com"),
        ModuleInstall('sseclient-py', 'pip', mname="sseclient",
                      purpose="A Python client for SSE event sources that seamlessly "
                      + "integrates with urllib3 and requests."),
        ModuleInstall('python-nest', 'pip', usage="IoT",
                      purpose="Python API and command line tool for talking to the Nest™ Thermostat"),
        ModuleInstall('VarEvents', 'pip', usage="IoT",
                      purpose="Python module to create variables that can raise custom events."),
        ModuleInstall('PyISY', 'pip', usage="IoT",
                      purpose="Python module to talk to ISY994 from UDI."),
        ModuleInstall('python-pushover', 'pip', usage="IoT",
                      purpose="Comprehensive bindings and command line utility for the Pushover notification service"),
        ModuleInstall('transmissionrpc', 'pip', usage="IoT",
                      purpose="Python module that implements the Transmission bittorent client RPC protocol."),
        ModuleInstall('pyowm', 'pip', usage="DATA",
                      purpose="A Python wrapper around the OpenWeatherMap web API"),
        ModuleInstall('sleekxmpp', 'pip', usage="IoT",
                      purpose="SleekXMPP is an elegant Python library for XMPP (aka Jabber, Google Talk, etc)."),
        ModuleInstall('dnspython3', 'pip', usage="IoT",
                      purpose="A DNS toolkit for Python 3.x"),
        ModuleInstall('blockchain', 'pip', usage="IoT",
                      purpose="Blockchain API library (v1)"),
        ModuleInstall('python-mpd2', 'pip', usage="IoT", mname="mdp2",
                      purpose="A Python MPD client library"),
        ModuleInstall('hikvision', 'pip', usage="IoT",
                      purpose="Provides a python interface to interact with a hikvision camera"),
        ModuleInstall('jsonrpc-requests', 'pip', mname="jsonrpc_requests",
                      purpose="A JSON-RPC client library, backed by requests"),
        ModuleInstall(
            'cookies', 'pip', purpose="Friendlier RFC 6265-compliant cookie parser/renderer"),
        ModuleInstall('responses', 'pip',
                      purpose="A utility library for mocking out the `requests` Python library."),
        ModuleInstall('python-forecastio', 'pip', usage="IoT",
                      purpose="A thin Python Wrapper for the Forecast.io weather API"),
        ModuleInstall('pyserial', 'pip', mname="serial",
                      usage="IoT", purpose="Python Serial Port Extension"),
        ModuleInstall('PyMata', 'pip', usage="IoT",
                      purpose="A Python Protocol Abstraction Library For Arduino Firmata"),
        ModuleInstall('pyRFXtrx', 'pip', mname="RFXtrx",
                      purpose="A Python library to communicate with the RFXtrx family of devices"),
        # ModuleInstall('pymysensors', 'github', 'theolind', mname="mysensors",
        # purpose="Python API for talking to a MySensors gateway"),
        ModuleInstall('pynetgear', 'pip', usage="IoT",
                      purpose="Access Netgear routers using their SOAP API"),
        ModuleInstall('netdisco', 'pip', usage="IoT",
                      purpose="Discover devices on your local network"),
        ModuleInstall('pywemo', 'pip', usage="IoT",
                      purpose="Access WeMo switches using their SOAP API"),
        ModuleInstall('python-wink', 'pip', mname="pywink",
                      purpose="Python implementation of the Wink API"),
        ModuleInstall('slacker', 'pip', purpose="Slack API client"),
        ModuleInstall(
            'vincenty', 'pip', purpose="Calculate the geographical distance between 2 points with extreme accuracy."),
        ModuleInstall('pyusb', 'pip', usage="IoT",
                      purpose="Python USB access module"),
        ModuleInstall('temperusb', 'pip', usage="IoT",
                      purpose="Reads temperature from TEMPerV1 devices (USB 0c45:7401)"),
        ModuleInstall('pyedimax', 'github', 'rkabadi', usage="IoT",
                      purpose="Pyedimax is a python library for interfacing with the Edimax Smart Plug switches SP-1101W and SP-2101W"),

        # Uncomment for Raspberry Pi
        # ModuleInstall('RPi.GPIO', 'pip'),
        # uncomment on a Raspberry Pi / Beaglebone
        # ModuleInstall('Adafruit_Python_DHT', 'github', 'mala-zaba', purpose="Adafruit temperature/humidity sensor", usage="IoT"),

        ModuleInstall('paho-mqtt', 'pip', mname="paho.mqtt", usage="IoT",
                      purpose="MQTT version 3.1/3.1.1 client class"),
        # ModuleInstall('pymodbus', 'github', 'bashwork', usage='IoT'),
        ModuleInstall('python-verisure', 'github', 'persandstrom', mname="verisure", usage='IoT',
                      purpose="A python module for reading and changing status of verisure devices through mypages."),
        ModuleInstall('voluptuous', 'pip',
                      purpose="Voluptuous is a Python data validation library"),
        ModuleInstall('astral', 'pip',
                      purpose="Calculations for the position of the sun and moon."),
        ModuleInstall('voluptuous_serialize', 'pip',
                      purpose="for homeassistant"),

        ModuleInstall('homeassistant', 'pip',
                      purpose="Home Assistant is a home automation platform running on Python 3. "
                      + "The goal of Home Assistant is to be able to track and control all devices at "
                      + "home and offer a platform for automating control.",
                      web="https://github.com/balloob/home-assistant/",
                      usage="IoT"),
    ]
    return [_ for _ in mod if _ is not None]



"""
@brief      test log(time=20s)
"""

import sys
import os
import unittest
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pycode import get_temp_folder
from pymyinstall.installhelper.module_install import ModuleInstall


class TestDownloadTables (unittest.TestCase):

    def test_install_tables(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")
        temp = get_temp_folder(__file__, "temp_download_tables")

        if sys.platform.startswith("win"):
            m = ModuleInstall("tables", "wheel", fLOG=fLOG)
            whl = m.download(temp_folder=temp, source="2")
            self.assertTrue(os.path.exists(whl))


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=20s)

skip this test for regular run
"""

import sys
import os
import unittest
import warnings
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pycode import is_travis_or_appveyor, skipif_circleci
from pymyinstall.installhelper import run_cmd


class TestScriptInstallCli(unittest.TestCase):

    @skipif_circleci("stuck")
    def test_script_help(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        if is_travis_or_appveyor() == "travis":
            # run_cmd no end on travis.
            return
        script = os.path.join(os.path.dirname(os.path.abspath(
            __file__)), "..", "..", "src", "pymyinstall", "cli", "pymy_install.py")
        if not os.path.exists(script):
            raise Exception(script)
        scriptu = os.path.join(os.path.dirname(os.path.abspath(
            __file__)), "..", "..", "src", "pymyinstall", "cli", "pymy_update.py")
        if not os.path.exists(script):
            raise Exception(script)

        exe = sys.executable

        cmd = exe + " " + script + " --help"
        out, err = run_cmd(cmd, wait=True, fLOG=fLOG)
        if "usage: pymy_install.py" not in out:
            raise Exception(out)

        cmd = exe + " " + scriptu + " --help"
        out, err = run_cmd(cmd, wait=True, fLOG=fLOG)
        if "usage: pymy_update.py" not in out:
            if is_travis_or_appveyor() == "appveyor":
                warnings.warn(
                    "CLI ISSUE cmd:\n{0}\nOUT:\n{1}\nERR\n{2}".format(cmd, out, err))
            else:
                raise Exception(
                    "cmd:\n{0}\nOUT:\n{1}\nERR\n{2}".format(cmd, out, err))

    @skipif_circleci("stuck")
    def test_script_schedule(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        if is_travis_or_appveyor() == "travis":
            # run_cmd no end on travis.
            return
        script = os.path.join(os.path.dirname(os.path.abspath(
            __file__)), "..", "..", "src", "pymyinstall", "cli", "pymy_install.py")
        self.assertTrue(os.path.exists(script))
        scriptu = os.path.join(os.path.dirname(os.path.abspath(
            __file__)), "..", "..", "src", "pymyinstall", "cli", "pymy_update.py")
        self.assertTrue(os.path.exists(script))

        exe = sys.executable

        cmd = exe + " " + script + " --schedule Cartopy --source=2"
        out, err = run_cmd(cmd, wait=True, fLOG=fLOG)
        fLOG(out)
        fLOG("---")
        fLOG(err)
        if "[install-check] ## flake8 ## [begin]" not in out:
            warnings.warn(
                "--CLI ISSUE cmd--\n{0}\n--OUT--\n{1}\n--ERR--\n{2}".format(cmd, out, err))

        cmd = exe + " " + scriptu + " --schedule --set=minimal --source=2"
        out, err = run_cmd(cmd, wait=True, fLOG=fLOG)
        if "[update-check] ## pycodestyle ## [begin]" not in out:
            if is_travis_or_appveyor() == "appveyor":
                warnings.warn(
                    "CLI ISSUE cmd:\n{0}\nOUT:\n{1}\nERR\n{2}".format(cmd, out, err))
            else:
                raise Exception(
                    "cmd:\n{0}\nOUT:\n{1}\nERR\n{2}".format(cmd, out, err))


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=80s)
"""
import unittest
import warnings
from pyquickhelper.loghelper import fLOG
from pymyinstall.installhelper.module_install_exceptions import MissingPackageOnPyPiException
from pymyinstall.installhelper.module_install_exceptions import MissingVersionOnPyPiException, AnnoyingPackageException
from pymyinstall.packaged import small_set


class TestAllPyPiModule(unittest.TestCase):

    def test_pipy_version(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        subset = {"cubehelix", "dataspyre", "ete3", "heapdict", "libpython",
                  "natgrid", "onedrive-sdk-python",
                  "orange3", "orange3-associate", "orange3-network", "orange3-text",
                  "py-earth", "pyexecjs", "pymc3", "pyreact", "pythonqwt", "qtpy"}

        mods = small_set()
        mods = [_ for _ in mods if _.name in subset]
        self._pipy_version(mods, nbmax=15)

    def test_all_pipy_version(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        mods = small_set()[:10]
        self._pipy_version(mods)

    def _pipy_version(self, mods, nbmax=15):
        error = []
        annoying = []
        for mod in mods:
            try:
                v = mod.get_pypi_version()
                fLOG(mod.name, " --> ", v)
                if v is None:
                    error.append((mod.name, "None", None))
            except MissingPackageOnPyPiException as e:
                error.append((mod.name, "pipy", e))
            except MissingVersionOnPyPiException as ee:
                error.append((mod.name, "version", ee))
            except AnnoyingPackageException as eee:
                annoying.append((mod.name, "?", eee))

        if len(error) > nbmax:
            # we accept some errors
            # joblib seems to give errors from time to time
            # multipledispatch
            # ipython --> jupyter (transitionning)
            raise MissingPackageOnPyPiException("Two many errors\n" +
                                                "\n".join("{0}:{1}\n   {2}".format(a, b, c) for a, b, c in sorted(error)))

        if len(annoying) > 0:
            ans = "\n".join(str(_) for _ in annoying)
            fLOG("Annoying\n", ans)
            warnings.warn("ANNOYING PACKAGES\n{0}".format(ans))


if __name__ == "__main__":
    unittest.main()



"""
@file
@brief Various functions to install some applications such as `7z <http://www.7-zip.org/>`_.
"""
from __future__ import print_function
import sys
import re
import os

from ..installhelper.install_cmd_helper import run_cmd, unzip_files
from .install_custom import download_page, download_file


def install_graphviz(temp_folder=".", fLOG=print, install=True,
                     force_download=False, version=None, source=None):
    """
    Install `Graphviz <http://www.graphviz.org/>`_.
    It does not do it a second time if it is already installed.

    @param      temp_folder     where to download the setup
    @param      fLOG            logging function
    @param      install         install (otherwise only download)
    @param      force_download  force the downloading of Graphviz
    @param      version         specify a version (unused)
    @param      source          change the source
    @return                     temporary file
    """
    if version is not None:
        raise ValueError("cannot specify a version")
    if sys.platform.startswith("win"):
        if source == "2":
            url = "http://www.xavierdupre.fr/enseignement/setup/"
            full = "graphviz-2.38.msi"
            url += full
            outfile = os.path.join(temp_folder, full)
            if not os.path.exists(temp_folder):
                os.makedirs(temp_folder)
        elif source == "zip":
            url = "http://www.xavierdupre.fr/enseignement/setup/"
            full = "Graphviz.zip"
            url += full
            outfile = os.path.join(temp_folder, full)
            if not os.path.exists(temp_folder):
                os.makedirs(temp_folder)
        else:
            link = "http://www.graphviz.org/Download_windows.php"
            repl = "http://www.graphviz.org"
            page = download_page(link)
            reg = re.compile("href=\\\"(.*?[.]msi)\\\"")
            alls = reg.findall(page)
            if len(alls) == 0:
                raise Exception(
                    "unable to find a link on a .msi file on page: " +
                    page)

            url = repl + alls[0]
            full = url.split("/")[-1]
            outfile = os.path.join(temp_folder, full)

        fLOG("[pymy] download ", url)
        local = download_file(url, outfile)
        if install:
            if source == "zip":
                fLOG("[pymy] dest ", os.path.abspath(temp_folder))
                unzip_files(local, temp_folder, fLOG=fLOG)
            else:
                run_cmd("msiexec /i " + local, fLOG=fLOG, wait=True)
        return local
    else:
        raise NotImplementedError("not available on platform " + sys.platform)



"""
@brief      test log(time=10s)
"""
import os
import unittest
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pycode import get_temp_folder
from pymyinstall import download_revealjs


class TestRevealjs(unittest.TestCase):

    def test_install_revealjs(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        temp = get_temp_folder(__file__, "temp_install_revealjs")
        dest = get_temp_folder(__file__, "temp_install_revealjs_dest")
        fs = download_revealjs(temp, dest, fLOG=fLOG)
        fLOG(fs)
        self.assertTrue(len(fs) > 0)
        for a in fs:
            self.assertTrue(os.path.exists(a))


if __name__ == "__main__":
    unittest.main()



import sys
print(sys.executable)
import platform
print(platform.architecture())
sys.path.append("src")
sys.path.append("../pyquickhelper/src")
print("----")
import pyquickhelper
from pymyinstall import win_python_setup
from pymyinstall.packaged import small_set


if True:
    win_python_setup(module_list=small_set(), verbose=True,
                     download_only=False,
                     no_setup=False,
                     selection={},
                     notebooks=[],
                     tutorial=["french"])



"""
@brief      test log(time=45s)
"""

import sys
import os
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall.installcustom import install_sqlitespy


class TestSqliteSpy (unittest.TestCase):

    def test_install_sqllitespy(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")
        fold = os.path.abspath(os.path.split(__file__)[0])
        temp = os.path.join(fold, "temp_sqllitespy")
        if not os.path.exists(temp):
            os.mkdir(temp)
        for _ in os.listdir(temp):
            if os.path.isfile(os.path.join(temp, _)):
                os.remove(os.path.join(temp, _))

        if sys.platform.startswith("win"):
            fLOG(
                "http://www.yunqa.de/delphi/lib/exe/fetch.php?hash=938481&media=" +
                "http%3A%2F%2Fwww.yunqa.de%2Fdelphi%2Fdownloads%2FSQLiteSpy_1.9.7.zip")
            exe = install_sqlitespy(temp, fLOG=fLOG, backup=True)
            fLOG("exe", exe)
            self.assertTrue(os.path.exists(exe))


if __name__ == "__main__":
    unittest.main()



# execute in admin mode
import sys
sys.path.append("../src")
from pymyinstall.packaged import update_all
update_all(temp_folder="build/update_modules", verbose=True,
           skip_module=[],
           source="2")



"""
@brief      test log(time=20s)
"""

import sys
import os
import unittest
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pycode import (
    is_travis_or_appveyor, get_temp_folder, skipif_appveyor)
from pymyinstall.installhelper.module_install import ModuleInstall
from pymyinstall.installhelper import compare_version


class TestDownloadNumpy (unittest.TestCase):

    @skipif_appveyor("skip numpy test")
    def test_install_numpy(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        if sys.platform.startswith("win") and sys.version_info[0] >= 3:
            temp = get_temp_folder(__file__, "temp_download_numpy")
            source = "2" if is_travis_or_appveyor() else None
            m = ModuleInstall(
                "numpy",
                "wheel",
                fLOG=fLOG,
                source=source)
            name = m.get_exewheel_url_link2(wheel=True, source=source)
            fLOG(m.existing_version)
            r = compare_version(m.existing_version, "1.10.1")
            self.assertTrue(r >= 0)
            self.assertTrue("unoptimized" not in name[0])
            whl = m.download(temp_folder=temp, source="2")
            fLOG(m.version)
            self.assertTrue(os.path.exists(whl))
            self.assertTrue("unoptimized" not in whl)


if __name__ == "__main__":
    unittest.main()



# execute in admin mode
import sys
sys.path.append("../src")
skip = ["actuariat_python", "code_beatrix", "ensae_teaching_cs",
        "jupytalk", "jyquickhelper", "mlstatpy",
        "pyquickhelper", "pyensae", "pyrsslocal", "pymyinstall",
        "pymmails", "teachpyx", 'ensae_projects',
        # 2017
        'mlprodict', 'lightmlrestapi', 'lightmlboard', 'pandas_streaaming',
        'pyenbc', 'mlinsights', 'cpyquickhelper', 'tkinterquickhelper',
        # 2018
        'manydataapi', 'ensae_traching_dl', 'sparkouille', 'csharpy', 'chsarpyml',
        # 2019
        'lecture_citation', 'skl2onnx', 'onnxmltoosl', 'onnxruntime', 'mathenjeu',
        'wrapclib', 'pymlbenchmark', '_benchmarks',
        ]
from pymyinstall.packaged import install_all
install_all(temp_folder="build/update_modules{0}{1}{2}".format(*sys.version_info[:3]),
            verbose=True, skip_module=skip, source="2")



"""
@file
@brief Functions to prepare a setup on Windows, R functions
"""

import json
import os
import sys

if sys.version_info[0] == 2:
    from codecs import open
    from StringIO import StringIO
else:
    from io import StringIO


r_kernel = {
    "display_name": "R%s",
    "language": "R",
    "argv": [
        "%R_HOME%\\bin\\x64\\R.exe",
        "--quiet",
        "-e",
        "IRkernel::main()",
        "--args",
        "{connection_file}"
    ]
}

python_kernel = {
    "language": "python",
    "display_name": "Python 3%s",
    "argv": [
        "%PYTHON_WINHOME%\\python.exe",
        "-m",
        "IPython.kernel",
        "-f",
        "{connection_file}"
    ]
}

julia_kernel = {
    "display_name": "Julia%s",
    "language": "julia",
    "argv": [
        "%JULIA_HOME%\\bin\\julia.exe",
        "-i",
        "-F",
        "%JULIA_PKGDIR%\\v0.3\\IJulia\\src\\kernel.jl",
        "{connection_file}"
    ],
    "codemirror_mode": "julia"
}


def add_kernel_jupyter(kernel, path, tools_path, python_path, suffix="WP"):
    """
    add a kernel to jupyter

    @param      kernel          dictionary (see global variable in this module)
    @param      path            where to add it
    @param      tools_path      tools paths
    @param      python_path     python path
    @param      suffix          added in bracket at the end of the display name
    @return                     created file
    """
    kernel = kernel.copy()
    suffix = " (%s)" % suffix
    kernel["display_name"] = kernel["display_name"] % suffix
    name = kernel["display_name"]
    name = name.replace(" ", "_").replace("(", "").replace(")", "")
    fold = os.path.join(path, name)
    if not os.path.exists(fold):
        os.mkdir(fold)
    st = StringIO()

    # replacements
    argv = kernel["argv"]
    for i in range(len(argv)):
        if "%" in argv[i]:
            argv[i] = argv[i].replace("%PYTHON_WINHOME%", python_path) \
                             .replace("%R_HOME%", os.path.join(tools_path, "R")) \
                             .replace("%JULIA_HOME%", os.path.join(tools_path, "Julia")) \
                             .replace("%JULIA_PKGDIR%", os.path.join(tools_path, "Julia", "pkg"))

    # dump
    if sys.version_info[0] == 2:
        json.dump(kernel, st, encoding="utf8")
    else:
        json.dump(kernel, st)
    ker = os.path.join(fold, "kernel.json")
    with open(ker, "w", encoding="utf8") as f:
        f.write(st.getvalue())
    return ker


def install_kernels(tools_path, python_path, suffix="WP"):
    """
    install available kernels on Windows

    @param      tools_path      tools paths
    @param      python_path     python path
    @param      suffix          suffix
    @return                     list of creating files
    """
    dest = os.environ["ProgramData"]
    jupyter = os.path.join(dest, "jupyter", "kernels")
    if not os.path.exists(jupyter):
        os.makedirs(jupyter)
    res = []
    if os.path.exists(python_path):
        res.append(
            add_kernel_jupyter(python_kernel, jupyter, tools_path, python_path, suffix))
    if os.path.exists(os.path.join(tools_path, "R")):
        res.append(
            add_kernel_jupyter(r_kernel, jupyter, tools_path, python_path, suffix))
    if os.path.exists(os.path.join(tools_path, "Julia")):
        res.append(
            add_kernel_jupyter(julia_kernel, jupyter, tools_path, python_path, suffix))
    return res



"""
@file
@brief Various helper about shortcuts and links
"""

import os
import sys
import platform

if sys.version_info[0] == 2:
    FileNotFoundError = Exception


def add_shortcut_to_desktop(file, name, description="", arguments="", icon=None, workdir=None):
    """
    @see fn add_shortcut
    """
    return add_shortcut(file, name, description, arguments, icon, workdir)


def add_shortcut(target, name, description="", arguments="",
                 icon=None, workdir=None, folder=None):
    """
    Adds a shortcut to the desktop.

    @param      target      target
    @param      name        name of the shortcut
    @param      description description
    @param      arguments   arguments
    @param      icon        icon
    @param      workdir     working directory
    @param      folder      where to save the shortcut (None for the desktop)
    @return                 path to the shortcut
    """
    if not sys.platform.startswith("win"):
        raise NotImplementedError(
            "not implemented on this platform: " +
            sys.platform)

    try:
        import winshell
    except ImportError as e:
        if "DLL load failed" in str(e):
            os.environ["PATH"] = os.environ[
                "PATH"] + ";" + os.path.split(sys.executable)[0] + r"\lib\site-packages\pywin32_system32"
            try:
                import winshell
            except ImportError:
                raise ImportError(
                    r"you should run the following in your current folder:\ncopy C:\%s\lib\site-packages\pywin32_system32\py*.dll %s" %
                    (os.path.split(
                        sys.executable),
                        os.getcwd())) from e
        else:
            raise e

    if folder is None:
        folder = winshell.desktop()
    link_filepath = os.path.join(folder, name + ".lnk")

    args = ["/c", "start", target, arguments]

    with winshell.shortcut(link_filepath) as link:
        link.path = r"%windir%\system32\cmd.exe"
        link.description = description if description is not None else name
        link.arguments = " ".join(args)
        if icon is not None:
            link.icon_location = (icon, 0)
        if workdir is not None:
            link.working_directory = workdir

    if not os.path.exists(link_filepath):
        raise FileNotFoundError(link_filepath)
    return link_filepath


def suffix():
    """
    add a suffix to a shortcut name = python version + architecture

    @return     string
    """
    ver = ".".join(str(_) for _ in sys.version_info[:2])
    arc = platform.architecture()[0]
    return "{0}.{1}".format(arc, ver)


if __name__ == "__main__":
    path = r"C:\github\pymyinstall\dist\win_python_setup"
    target = r"tools\R\bin\R.exe"
    name = "R Console"
    icon = r"tools\icons\r.ico"
    add_shortcut(target=target, folder=path, icon=icon, name=name)

    path = r"C:\github\pymyinstall\dist\win_python_setup"
    target = r"python\python.exe"
    name = "Python Console"
    icon = r"tools\icons\python.ico"
    add_shortcut(target=target, folder=path, icon=icon, name=name)



"""
@brief      test log(time=1s)
"""
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall.installhelper import missing_dependencies


class TestMissingModules (unittest.TestCase):

    def test_missing_module(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        deps = missing_dependencies()
        self.assertTrue(isinstance(deps, dict))
        for k, v in sorted(deps.items()):
            fLOG(k, "--->", ", ".join(v))


if __name__ == "__main__":
    unittest.main()



#!python
"""
script which looks into the dependencies for a module

.. todoext::
    :title: add a script to look into dependencies
    :tag: done
    :cost: 0.5
    :date: 2016-09-18
    :hidden:
    :release: 1.1

    Add a script to check the missing dependencies.

.. versionadded:: 1.1
"""
from __future__ import print_function
import sys
import os
import argparse


def get_parser():
    """
    defines the way to parse the script ``pymy_deps``
    """
    parser = argparse.ArgumentParser(
        description='look dependencies for modules')
    parser.add_argument(
        'module',
        nargs="*",
        default="all",
        help='update only the list of modules included in this list or all modules if not specified or equal to all')
    return parser


def do_main(list_modules=None):
    """
    calls function @see fn missing_dependencies

    @param      list_modules     list of modules to update or None for all
    """
    try:
        from pymyinstall.installhelper import missing_dependencies, get_module_dependencies
    except ImportError:
        pfolder = os.path.normpath(os.path.join(
            os.path.abspath(os.path.dirname(__file__)), "..", ".."))
        sys.path.append(pfolder)
        if "pymyinstall" in sys.modules:
            del sys.modules["pymyinstall"]
        if "pymyinstall.installhelper" in sys.modules:
            del sys.modules["pymyinstall.installhelper"]
        try:
            from pymyinstall.installhelper import missing_dependencies, get_module_dependencies
        except ImportError as e:
            mes = "pfolder={0}\nPATH:\n{1}".format(
                pfolder, "\n".join(sys.path))
            raise ImportError(mes) from e
    res = missing_dependencies(list_modules)
    if len(res) > 0:
        for r in res:
            print(r)
    elif list_modules is not None and len(list_modules) > 0:
        for mod in list_modules:
            print("checking ", mod)
            res = get_module_dependencies(mod)
            for r, v in sorted(res.items()):
                print("  ", r, v)
    else:
        print("no missing dependencies")


def main():
    """
    calls function @see fn missing_dependencies but is meant to be added to scripts folder,
    parse command line arguments
    """
    parser = get_parser()
    try:
        res = parser.parse_args()
    except SystemExit:
        print(parser.format_usage())
        res = None

    if res is not None:
        list_module = None if res.module in [
            "all", "", "-", None, []] else res.module
        do_main(list_modules=list_module)


if __name__ == "__main__":
    main()



"""
@file
@brief Functions to remember a step was done, uses files.
"""
import os
import datetime
import sys

if sys.version_info[0] == 2:
    from codecs import open


def mark_step(folder, step_name, content=""):
    """
    create a file to remember was done (running it again takes times)

    @param      folder      folder where to write
    @param      step_name   step_name
    @param      content     what to write in the file
    @return                 created file name
    """
    name = os.path.join(folder, "log.step.{0}.txt".format(step_name))
    with open(name, "w", encoding="utf8") as f:
        f.write("# -- STEP {0} DONE --\n\n".format(step_name))
        f.write("# -- {0}\n\n".format(datetime.datetime.now()))
        if content is not None:
            f.write(content)
        f.write("\n# -- {0}".format(datetime.datetime.now()))
    return name


def is_step_done(folder, step_name):
    """
    checks a file was written with function @see fn mark_step

    @param      folder      folder where to write
    @param      step_name   step_name
    @return                 boolean
    """
    name = os.path.join(folder, "log.step.{0}.txt".format(step_name))
    if not os.path.exists(name):
        return False
    s = "# -- STEP {0} DONE --\n\n".format(step_name)
    with open(name, "r", encoding="utf8") as f:
        content = f.read()
    if content.startswith(s):
        with open(name, "a", encoding="utf8") as f:
            f.write("\n# skip {0}".format(datetime.datetime.now()))
        return True
    else:
        return False



"""
@file
@brief Functions to prepare a setup on Windows, R functions
"""
import sys
import os
from ..installhelper.install_cmd_helper import run_cmd

if sys.version_info[0] == 2:
    FileNotFoundError = Exception


_script = os.path.join(
    os.path.abspath(os.path.dirname(__file__)), "R_install.r")


class RBatchException(Exception):

    """
    raised when running R in batch mode
    """
    pass


def r_run_script(r_path, script, output=None):
    """
    run a script on R

    @param      r_path      r location
    @param      script      script to run
    @param      output      where to store the output (can be None)
    @return                 output
    """
    exe = os.path.join(r_path, "bin", "x64", "R.exe")
    if not os.path.exists(exe):
        raise FileNotFoundError(exe)

    os.environ["R_LIBS"] = os.path.join(r_path, "library")
    cmd = [exe, "CMD", "BATCH", script]
    if output is not None:
        cmd.append(output)
    cmd = " ".join(cmd)
    out, err = run_cmd(cmd, wait=True)
    if err is not None and len(err) > 0:
        raise RBatchException(
            "CMD:\n{0}\nOUT:\n{1}\nERR--G:\n{2}".format(cmd, out, err))
    return out


def get_package_description(r_path, pack):
    """
    returns the description of an R package as a dictionary

    @param      r_path      path to R
    @param      pack        package name
    @return                 dictionary
    """
    path = os.path.join(r_path, "library", pack)
    if not os.path.exists(path):
        raise FileNotFoundError(path)
    version = os.path.join(path, "DESCRIPTION")
    if not os.path.exists(version):
        raise FileNotFoundError(version)
    with open(version, "r") as f:
        lines = f.readlines()
    res = {}
    for line in lines:
        spl = line.split(":")
        if len(spl) > 1:
            key = spl[0]
            val = ":".join(spl[1:])
            res[key] = val.strip("\r\n\t ")
    return res



"""
@file
@brief Functions to get a status on the distribution

.. versionadded:: 1.1
"""
import os

import time
from .module_install_version import get_pypi_version
from ..packaged import small_set
from ..packaged.packaged_config import classifiers2string
from .pip_helper import get_installed_distributions


def get_installed_modules(pypi=False, skip_betas=False, fLOG=None, stop=-1, short_list=None):
    """
    Returns all modules recored in this modules.
    Adds installed modules.

    @param      pypi            add pypi version
    @param      skip_betas      skip the intermediate functions
    @param      stop            stop after *stop* (or -1 for all)
    @param      short_list      short list of modules or None for all
    @param      fLOG            logging function
    @return                     list of dictionaries
    """
    mod = small_set()
    mod.sort()
    rows = [_.as_dict(rst_link=True) for _ in mod]
    for row in rows:
        row["classifier"] = classifiers2string(row["classifier"])
    keys = {mod["name"].lower(): mod for mod in rows}
    keys.update({mod["mname"].lower(): mod for mod in rows if mod["mname"]})

    all_installed = []
    dists = get_installed_distributions()
    if short_list:
        dists = [_ for _ in dists if _.project_name in short_list]
    pack = map(lambda d: (d.project_name.lower(), d), dists)
    for i, (_, dist) in enumerate(sorted(pack)):
        if i >= stop >= 0:
            break
        res = {}
        key = dist.project_name.lower()
        if key in keys:
            res.update(keys[key])

        names = [dist.key, dist.project_name, dist.key.replace("-", "_"),
                 res.get("mname", None), res.get("name", None)]
        date = ""
        location = None
        for name in names:
            if name is None:
                continue
            ini = os.path.join(dist.location, name, "__init__.py")
            if os.path.exists(ini):
                date = time.ctime(os.path.getctime(ini))
                location = os.path.dirname(ini)
                break
            ini = os.path.join(dist.location, name + ".py")
            if os.path.exists(ini):
                date = time.ctime(os.path.getctime(ini))
                location = ini
                break
        res["installed"] = dist.project_name
        if ini is not None:
            res["location"] = location
            res["installed_date"] = date
        res["installed_version"] = dist.version

        # update?
        if pypi:
            available = get_pypi_version(
                dist.project_name, skip_betas=skip_betas, full_list=True)

            if not available:
                msg = '-'
            elif available[0] != dist.version:
                msg = '!='
            else:
                msg = "=="
            res["available"] = available
            res["pypi"] = available[0]
        else:
            msg = ''
            available = None

        res["diff_pypi"] = msg

        if fLOG:
            sdist = str(dist.version)
            pkg_info = '{dist.project_name:30} {sdist:20}'.format(
                dist=dist, sdist=sdist)
            key = key if dist.key != dist.project_name else ""
            mes = '{pkg_info} {msg:3} - {date:20} --- {available} - {location}'
            mes = mes.format(pkg_info=pkg_info, msg=msg, date=date, dist=dist,
                             available=available, location=res.get('location', ''))
            fLOG("[pymy] {0}/{1} - {2}".format(i, len(dists), mes))

        all_installed.append(res)
    return all_installed



# -*- coding: utf-8 -*-
"""
@file
@brief Defines a set of modules for more machine learning or student projects.
"""
from ..installhelper.module_install import ModuleInstall


def cloud_set():
    """
    modules introduced by students or needed for student projects, it requires the modules in set *extended*
    """
    mod = [
        ModuleInstall("botocore", "pip", usage="AWS",
                      purpose="A low-level interface to a growing number of Amazon Web Services. " +
                      "The botocore package is the foundation for the AWS CLI as well as boto3."),
        ModuleInstall("s3transfer", "pip", usage="AWS",
                      purpose="An Amazon S3 Transfer Manager"),
        ModuleInstall("boto3", "pip", usage="AWS",
                      purpose="A Python interface to Amazon Web Services"),
        ModuleInstall("s3fs", "pip", usage="AWS",
                      purpose="Convenient Filesystem interface over S3"),
        ModuleInstall("boto", "pip",
                      purpose="Amazon Web Services Library"),
        ModuleInstall("google-auth-httplib2", "pip", mname="google_auth_httplib2",
                      purpose="This library provides an httplib2 transport for google-auth."),
        ModuleInstall("google-auth", "pip", mname="google_auth",
                      purpose="This library simplifies using Google’s various server-to-server " +
                      "authentication mechanisms to access Google APIs."),
        ModuleInstall("google-api-python-client", "pip", mname="googleapiclient",
                      purpose="The Google API Client for Python is a client library for accessing the Plus, " +
                      "Moderator, and many other Google APIs."),
        ModuleInstall("googlemaps", "pip",
                      purpose="Python client library for Google Maps API Web Services"),
        ModuleInstall("python-gmaps", "pip", mname="gmaps",
                      purpose="Google Maps API client http://python-gmaps.readthedocs.org"),
    ]

    mod.append(ModuleInstall("adal", "pip",
                             purpose="The ADAL for Python library makes it easy for python application to authenticate " +
                             "to Azure Active Directory (AAD) in order to access AAD protected web resources."))
    mod.append(ModuleInstall("msrest", "pip",
                             purpose="AutoRest swagger generator Python client runtime."))
    mod.append(ModuleInstall("msrestazure", "pip",
                             purpose="AutoRest swagger generator Python client runtime. Azure-specific module."))
    for name in ['azure-nspkg',
                 'azure-common',
                 'azure-mgmt-nspkg',
                 'azure-mgmt-authorization',
                 'azure-mgmt-common',
                 'azure-storage',
                 'azure-mgmt-batch',
                 'azure-mgmt-cdn',
                 'azure-mgmt-cognitiveservices',
                 'azure-mgmt-commerce',
                 'azure-mgmt-compute',
                 'azure-mgmt-logic',
                 'azure-graphrbac',
                 'azure-mgmt-network',
                 'azure-mgmt-notificationhubs',
                 'azure-mgmt-powerbiembedded',
                 'azure-mgmt-redis',
                 'azure-mgmt-resource',
                 'azure-mgmt-scheduler',
                 'azure-mgmt-storage',
                 'azure-mgmt-web',
                 'azure-batch',
                 'azure-servicebus',
                 'azure-servicemanagement-legacy',
                 'azure-mgmt',
                 # addition 2017-05
                 "azure-keyvault",
                 "azure-datalake-store",
                 "azure-servicefabric",
                 "azure-mgmt-devtestlabs",
                 "azure-mgmt-documentdb",
                 "azure-mgmt-containerregistry",
                 "azure-mgmt-keyvault",
                 "azure-mgmt-dns",
                 "azure-mgmt-datalake-analytics",
                 "azure-mgmt-datalake-nspkg",
                 "azure-mgmt-trafficmanager",
                 "azure-mgmt-rdbms",
                 "azure-mgmt-datalake-store",
                 "azure-mgmt-iothub",
                 "azure-mgmt-sql",
                 "azure-mgmt-monitor",
                 # addition 2018-02
                 'azure_storage_common',
                 'azure_cosmosdb_nspkg',
                 'azure_cosmosdb_table',
                 'azure_eventgrid',
                 'azure_mgmt_advisor',
                 'azure_mgmt_applicationinsights',
                 'azure_mgmt_batchai',
                 'azure_mgmt_billing',
                 'azure_mgmt_consumption',
                 'azure_mgmt_containerinstance',
                 'azure_mgmt_containerservice',
                 'azure_mgmt_cosmosdb',
                 'azure_mgmt_datafactory',
                 'azure_mgmt_eventgrid',
                 'azure_mgmt_eventhub',
                 'azure_mgmt_hanaonazure',
                 'azure_mgmt_iothubprovisioningservices',
                 'azure_mgmt_loganalytics',
                 'azure_mgmt_machinelearningcompute',
                 'azure_mgmt_managementpartner',
                 'azure_mgmt_marketplaceordering',
                 'azure_mgmt_media',
                 'azure_mgmt_msi',
                 'azure_mgmt_recoveryservices',
                 'azure_mgmt_recoveryservicesbackup',
                 'azure_mgmt_relay',
                 'azure_mgmt_reservations',
                 'azure_mgmt_search',
                 'azure_mgmt_servermanager',
                 'azure_mgmt_servicebus',
                 'azure_mgmt_servicefabric',
                 'azure_mgmt_subscription',
                 'azure_storage_blob',
                 'azure_storage_file',
                 'azure_storage_queue',
                 'azure-storage-nspkg',
                 # addition 2019-01
                 'azure_loganalytics',
                 'azure_applicationinsights',
                 'azure_mgmt_iotcentral',
                 'azure_mgmt_datamigration',
                 'azure_mgmt_maps',
                 'azure_mgmt_policyinsights',
                 'azure_mgmt_managementgroups',
                 'azure_mgmt_devspaces',
                 'azure_mgmt_signalr',
                 ]:

        # azure part
        mname = name.replace("-", ".").replace("_", ".")
        if mname in ("azure.nspkg", "azure.mgmt.nspkg",
                     "azure.servicemanagement.legacy"):
            skip_import = True
        else:
            skip_import = False
        if mname == name:
            mname = None
        m = ModuleInstall(
            name, "pip", mname=mname, pip_options=["--pre"],
            purpose="Python wrapper for Azure API (HDInsight, Blog Storage)", usage="AZURE",
            skip_import=skip_import)
        mod.append(m)

    mod.append(ModuleInstall("azureml", "pip",
                             purpose="Microsoft Azure Machine Learning Python client library"))

    return [_ for _ in mod if _ is not None]



"""
@file
@brief Shortcuts to packaged
"""

from .config_helper import is_64bit
from .automate_install import update_all, install_all, find_module_install, install_module_deps
from .automate_install import install_module, update_module, download_module
from .packaged_config import minimal_set, small_set
from .packaged_config import get_package_set, name_sets_dataframe, classifiers2string



# -*- coding: utf-8 -*-
"""
@file
@brief Helpers about configurations
"""
import sys


def is_64bit():
    """
    tells if the python is 64 bit or not

    @return     boolean
    """
    return sys.maxsize > 2 ** 32



"""
@file
@brief Helpers, inspired from `utils.py <https://github.com/winpython/winpython/blob/master/winpython/utils.py>`_
"""


class WinInstallException(Exception):

    """
    exception raised by this package
    """
    pass


class WinInstallPackageException(Exception):

    """
    exception raised by this package when installing a Python package
    """
    pass


class WinInstallMissingDependency(Exception):
    """
    raised when a dependency is missing
    """
    pass


class WinInstallDistributionError(Exception):
    """
    raised when an issue is detected in the distribution
    """
    pass



"""
@brief      test log(time=120s)
"""

import sys
import os
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall.packaged import find_module_install


class TestLONGDownloadCntk(unittest.TestCase):

    @unittest.skipIf(sys.version_info[:2] == (3, 7), reason="not released yet on python 3.7")
    def test_install_cntk(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")
        fold = os.path.abspath(os.path.split(__file__)[0])
        temp = os.path.join(fold, "temp_download_cntk")
        if not os.path.exists(temp):
            os.mkdir(temp)
        for _ in os.listdir(temp):
            if os.path.isfile(os.path.join(temp, _)):
                os.remove(os.path.join(temp, _))

        if sys.platform.startswith("win"):
            fLOG("install", "cntk")
            m = find_module_install("cntk")
            m.fLOG = fLOG
            exe = m.download(temp_folder=temp)
            assert os.path.exists(exe)
            assert os.stat(exe).st_size > 100000


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=20s)
"""
import os
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall.installhelper.module_install import ModuleInstall


class TestDownload (unittest.TestCase):

    def test_install(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")
        fold = os.path.abspath(os.path.split(__file__)[0])
        temp = os.path.join(fold, "temp_download")
        if not os.path.exists(temp):
            os.mkdir(temp)
        for _ in os.listdir(temp):
            if os.path.isfile(os.path.join(temp, _)):
                os.remove(os.path.join(temp, _))
        if os.path.exists(os.path.join(temp, "jsdifflib-master")):
            for _ in os.listdir(os.path.join(temp, "jsdifflib-master")):
                os.remove(
                    os.path.join(
                        os.path.join(
                            temp,
                            "jsdifflib-master"),
                        _))

        m = ModuleInstall("jsdifflib", "github", gitrepo="cemerick", fLOG=fLOG)
        files = m.download(temp_folder=temp, unzipFile=True, source="2")
        self.assertTrue(len(files) > 0)
        for _ in files:
            self.assertTrue(os.path.exists(_))


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=1290s)

skip this test for regular run
"""

import sys
import os
import unittest
from pyquickhelper.loghelper import fLOG, CustomLog
from pyquickhelper.pycode import get_temp_folder
from pymyinstall.installcustom import install_python


class TestDownloadPythonMinimal(unittest.TestCase):

    def test_install_python_minimal(self):
        fLOG(__file__, self._testMethodName, OutputPrint=True)
        vers = "%d%d" % sys.version_info[:2]
        temp = get_temp_folder(
            __file__, "temp_py%s_minimal" % vers, clean=True, persistent=True)
        down = get_temp_folder(
            __file__, "temp_py%s_minimal_download" % vers, clean=True, persistent=True)
        self.assertEqual(len(os.listdir(down)), 0)

        fLOG("BEGIN in", temp)
        clog = CustomLog(temp)
        install_python(install=True, temp_folder=temp,
                       fLOG=clog, modules="minimal", custom=True, latest=True,
                       download_folder=temp + "_download")
        fLOG("END")
        if sys.platform.startswith("win"):
            pyt = os.path.join(temp, "python.exe")
            pip = os.path.join(temp, "Scripts", "pip.exe")
            if not os.path.exists(pyt):
                raise FileNotFoundError(pyt)
            if not os.path.exists(pip):
                raise FileNotFoundError(pip)
            post = os.path.join(temp, "Scripts", "pywin32_postinstall.py")
            with open(post, "r") as f:
                content = f.read()
            if "os.path.exists(dest)" not in content:
                raise Exception(
                    "'os.path.exists(dest)' not found in '{0}'".format(post))
        else:
            # already checked
            pass


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
@file
@brief Functions to prepare a setup on Windows

.. todoext::
    :title: enables js extension on jupyter
    :tag: enhancement

    Run something like::

        jupyter nbextension enable --py --sys-prefix widgetsnbextension
"""
from __future__ import print_function
import sys
import os

from ..installhelper.install_cmd_helper import run_cmd, get_pip_program
from .win_exception import WinInstallException

if sys.version_info[0] == 2:
    FileNotFoundError = Exception


def ipython_create_profile(config_path, python_path, name="win_profile", fLOG=print):
    """
    creates a ipython profile

    @param      config_path     where to create it
    @param      python_path     python path (to get ipython)
    @param      profile         name
    @param      fLOG            logging function
    @return                     profile path
    """
    if sys.platform.startswith("win"):
        ipython_path = os.path.join(python_path, "ipython.exe")
        if not os.path.exists(ipython_path):
            ipython_path = os.path.join(
                python_path, "Scripts", "ipython.exe")
            if not os.path.exists(ipython_path):
                raise FileNotFoundError(ipython_path)
    else:
        ipython_path = os.path.join(
            python_path, "ipython")

    cmd = " profile create {1} --ipython-dir={0}".format(config_path, name)
    cmd = ipython_path + cmd
    out, err = run_cmd(cmd, wait=True, fLOG=fLOG, change_path=python_path)
    profile = os.path.join(
        config_path, "profile_" + name, "ipython_config.py")
    if not os.path.exists(profile):
        raise WinInstallException(
            "missing file, unable to execute:\nFILE:\n{3}\nCMD:\n{0}\nOUT:\n{1}\nERR-6:\n{2}".format(cmd, out, err, profile))
    return os.path.dirname(profile)


def ipython_update_profile(profile_path):
    """
    update the profile with custom settings (file filters)

    @param      profile_path        path to profile
    """
    profile = os.path.join(profile_path, "ipython_kernel_config.py")
    with open(profile, "r") as f:
        content = f.read()
    add = """
                c.ContentsManager.hide_globs = ['__pycache__', '*.pyc', '*.pyo', '.DS_Store', '*.so', '*.dylib', '*~', ".ipynb_checkpoints",
                    ".kernel-*.json", ".kernel", ".RData", ".RHistory"]
                c.FileContentsManager.hide_globs = ['__pycache__', '*.pyc', '*.pyo', '.DS_Store',
                    '*.so', '*.dylib', '*~', ".ipynb_checkpoints", ".kernel-*.json", ".kernel", ".RData", ".RHistory"]
                """.split("\n")
    content += "\n".join(_.strip() for _ in add)
    with open(profile, "w") as f:
        f.write("\n" + content + "\n")


def install_jupyter_extension(python_path):
    """
    install jupyter extension

    @param  python_path     python path
    @return                 out, err
    """
    if sys.platform.startswith("win"):
        pip = os.path.join(python_path, "Scripts", "pip.exe")
        if not os.path.exists(pip):
            pip = get_pip_program(exe=python_path)
    else:
        pip = get_pip_program(exe=python_path)

    cmd = "{0} install https://github.com/ipython-contrib/IPython-notebook-extensions/archive/master.zip --user".format(
        pip)
    out, err = run_cmd(cmd)
    if err:
        raise WinInstallException(
            "unable to install jupyter extension\nOUT:{0}\nERR-7{1}".format(out, err))
    return out, err



"""
@brief      test tree node (time=3s)
"""
import os
import unittest
import warnings
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pycode import get_temp_folder
from pymyinstall.installhelper.install_venv_helper import create_virtual_env, run_venv_script


class TestReadme(unittest.TestCase):

    def test_venv_docutils08_readme(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        fold = os.path.dirname(os.path.abspath(__file__))
        readme = os.path.join(fold, "..", "..", "README.rst")
        assert os.path.exists(readme)
        with open(readme, "r", encoding="utf8") as f:
            content = f.read()

        temp = get_temp_folder(__file__, "temp_readme")

        if __name__ != "__main__":
            warnings.warn("does not work well from a virtual environment")
            return

        create_virtual_env(temp, fLOG=fLOG, packages=["docutils==0.8"])
        outfile = os.path.join(temp, "conv_readme.html")

        script = ["from docutils import core",
                  "import io",
                  'from docutils.readers.standalone import Reader',
                  'from docutils.parsers.rst import Parser',
                  'from docutils.parsers.rst.directives.images import Image',
                  'from docutils.parsers.rst.directives import _directives',
                  'from docutils.writers.html4css1 import Writer',
                  "from docutils.languages import _languages",
                  "from docutils.languages import en, fr",
                  "_languages['en'] = en",
                  "_languages['fr'] = fr",
                  "_directives['image'] = Image",
                  "with open('{0}', 'r', encoding='utf8') as g: s = g.read()".format(
                      readme.replace("\\", "\\\\")),
                  "settings_overrides = {'output_encoding': 'unicode', 'doctitle_xform': True, ",
                  "                      'initial_header_level': 2, 'warning_stream': io.StringIO()}",
                  "parts = core.publish_parts(source=s, parser=Parser(), reader=Reader(), source_path=None, destination_path=None,"
                  ",                          writer=Writer(), settings_overrides=settings_overrides)",
                  "with open('{0}', 'w', encoding='utf8') as f: f.write(parts['whole'])".format(
                      outfile.replace("\\", "\\\\")),
                  ]

        file_script = os.path.join(temp, "testreadme.py")
        with open(file_script, "w") as f:
            f.write("\n".join(script))

        run_venv_script(temp, file_script, fLOG=fLOG, file=True)
        with open(outfile, "r", encoding="utf8") as h:
            content = h.read()

        if "System Message" in content:
            raise Exception(content)


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=23s)
"""
import sys
import os
import unittest
import warnings
import subprocess
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pycode import get_temp_folder
from pyquickhelper.pycode import is_travis_or_appveyor
from pymyinstall.installhelper.install_cmd_helper import run_cmd


class TestPyMyInstallCli(unittest.TestCase):

    def test_install_set_schedule(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        this = os.path.abspath(os.path.dirname(__file__))
        script = os.path.normpath(os.path.join(
            this, "..", "..", "src", "pymyinstall", "cli", "pymy_install.py"))
        cmd = "{0} -u {1} {2}".format(
            sys.executable, script, "--set=pyquickhelper --schedule")
        try:
            out, err = run_cmd(cmd, wait=True, fLOG=fLOG,
                               communicate=True, timeout=60)
        except subprocess.CalledProcessError as e:
            mes = "CMD\n{0}\nOUT\n{1}\nERR\n{2}".format(
                e.cmd, e.output, e.stderr.read() if e.stderr else "")
            raise Exception(mes) from e
        if len(out) == 0:
            if is_travis_or_appveyor() == "appveyor":
                warnings.warn(
                    "CLI ISSUE cmd:\n{0}\nOUT:\n{1}\nERR\n{2}".format(cmd, out, err))
            else:
                raise Exception(
                    "cmd:\n{0}\nOUT:\n{1}\nERR\n{2}".format(cmd, out, err))

    def test_install_download(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        temp = get_temp_folder(__file__, "temp_install_download")
        this = os.path.abspath(os.path.dirname(__file__))
        script = os.path.normpath(os.path.join(
            this, "..", "..", "src", "pymyinstall", "cli", "pymy_install.py"))
        cmd = "{0} {1} {2} --force --folder={3}".format(
            sys.executable, script, "colorama xlrd --download", temp)
        out, err = run_cmd(cmd, wait=True, fLOG=fLOG,
                           communicate=True, timeout=60)
        fLOG("----", cmd)
        fLOG(out.replace("\r", "").replace("\n\n", "\n"))
        fLOG("-----")
        fLOG(err.replace("\r", "").replace("\n\n", "\n"))
        if len(out) == 0 and is_travis_or_appveyor() == "appveyor":
            warnings.warn(
                "CLI ISSUE cmd:\n{0}\nOUT:\n{1}\nERR\n{2}".format(cmd, out, err))
            return
        if "downloaded modules" not in out:
            raise Exception(
                "CMD: {0}\nOUT:\n{1}\nERR\n{2}".format(cmd, out, err))
        content = os.listdir(temp)
        if len(content) != 2:
            if is_travis_or_appveyor() == "appveyor":
                warnings.warn(
                    "CLI ISSUE cmd:\n{0}\nOUT:\n{1}\nERR\n{2}".format(cmd, out, err))
            else:
                raise Exception(
                    "cmd:\n{0}\nOUT:\n{1}\nERR\n{2}".format(cmd, out, err))


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=1s)
"""
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall.installhelper.module_install import compare_version
from pymyinstall.installhelper.module_install_version import get_module_version


class TestInstallModule (unittest.TestCase):

    def test_module_version(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")
        d = get_module_version(None)
        if len(d) < 10:
            for k, v in sorted(d.items()):
                fLOG(k, v)
            raise Exception("el {0}\nD\n{1}".format(len(d), d))

    def test_compare_version(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")
        self.assertEqual(compare_version("0.16.0", "0.16.2"), -1)


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=20s)
"""

import sys
import os
import unittest
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pycode import get_temp_folder
from pymyinstall.fix import fix_win32ctypes_core_cffi__advapi32_py


class TestFixEol(unittest.TestCase):

    def test_fix_eol(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        if sys.platform.startswith("win") and sys.version_info[:2] >= (3, 6):
            temp = get_temp_folder(__file__, "temp_fix_eol")
            data = os.path.join(temp, "..", "data", "_file.pytry")
            out = os.path.join(temp, "out.py")
            fix_win32ctypes_core_cffi__advapi32_py(data, out, fLOG=fLOG)
            self.assertTrue(os.path.exists(out))


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
@file
@brief Helpers, inspired from `utils.py <https://github.com/winpython/winpython/blob/master/winpython/utils.py>`_
"""
from __future__ import print_function

import sys
import os
import os.path as osp
import subprocess
import tarfile
import zipfile
import fnmatch
# import winreg
import glob
import shutil
from ..installcustom.install_custom_python import unzip7_files

from .win_helper import patch_shebang_line

if sys.version_info[0] == 2:
    FileNotFoundError = Exception


def is_program_installed(basename):
    """
    Return program absolute path if installed in PATH
    Otherwise, return None

    @param      basename        base name
    @return                     boolean
    """
    if os.path.exists(basename):
        return True
    for path in os.environ["PATH"].split(os.pathsep):
        abspath = osp.join(path, basename)
        if osp.isfile(abspath):
            return abspath


def extract_msi(fname, targetdir=None, verbose=False, fLOG=print):
    """
    Extract .msi installer to a temporary directory (if targetdir
    is None). Return the temporary directory path

    @param  fname           local installer (exe)
    @param  targetdir       where to install
    @param  verbose         verbose
    @param  fLOG            logging function
    @return                 targetdir
    """
    extract = 'msiexec.exe'
    bname = osp.basename(fname)
    args = ['/a', '%s' % bname]
    if not verbose:
        args += ['/quiet', '/qn']
    args += ['TARGETDIR=%s' % targetdir]
    cmd = [extract] + args
    fLOG("[pymy] RUN: ", cmd)
    subprocess.call(cmd, cwd=osp.dirname(fname))
    fLOG('fname=%s' % fname)
    fLOG('TARGETDIR=%s' % targetdir)

    if "python-3" in fname.lower():
        subprocess.call(
            [r'%s\%s' % (targetdir, 'python.exe'), '-m', 'ensurepip'],
            cwd=osp.dirname(r'%s\%s' % (targetdir, 'pythons.exe')))
        # We patch ensurepip live (shame) !!!!
        # rational: https://github.com/pypa/pip/issues/2328
        for fname in glob.glob(r'%s\Scripts\*.exe' % targetdir):
            patch_shebang_line(fname)
    return targetdir


def extract_exe(fname, targetdir=None, verbose=False, fLOG=print, szip="7z.exe"):
    """
    Extract *.exe* archive to a temporary directory (if targetdir
    is None). Return the temporary directory path

    @param  fname           local installer (exe)
    @param  targetdir       where to install
    @param  verbose         verbose
    @param  szip            path to 7z.exe
    @param  fLOG            logging function
    @return                 targetdir
    """
    extract = szip
    if not is_program_installed(extract):
        pp = os.environ["PATH"].split(";")
        raise FileNotFoundError(
            "Required program '%s' was not found in PATH:\n  %s" % (extract, "\n  ".join(pp)))
    bname = osp.basename(fname)
    args = ['x', '-o%s' % targetdir, '-aos', bname]
    if verbose:
        retcode = subprocess.call([extract] + args, cwd=osp.dirname(fname))
    else:
        p = subprocess.Popen([extract] + args, cwd=osp.dirname(fname),
                             stdout=subprocess.PIPE)
        p.communicate()
        p.stdout.close()
        retcode = p.returncode
    if retcode != 0:
        raise RuntimeError("Failed to extract %s (return code: %d),\nCMD: %s\nCMD*: %s\nin %s"
                           % (fname, retcode, str([extract] + args), " ".join([extract] + args), targetdir))
    return targetdir


def extract_archive(fname, targetdir=None, verbose=False, fLOG=print):
    """
    Extract .zip, .exe (considered to be a zip archive) or .tar.gz archive
    to a temporary directory (if targetdir is None).
    Return the temporary directory path

    @param  fname           zip file or exe file
    @param  targetdir       where to install
    @param  verbose         verbose
    @param  fLOG            logging function
    @return                 targetdir
    """
    if osp.splitext(fname)[1] in ('.zip', '.exe'):
        obj = zipfile.ZipFile(fname, mode="r")
    elif osp.splitext(fname)[1] in ('.7z'):
        unzip7_files(fname, fLOG=fLOG, dest=targetdir)
        return targetdir
    elif fname.endswith('.tar.gz'):
        obj = tarfile.open(fname, mode='r:gz')
    else:
        raise RuntimeError("Unsupported archive filename %s" % fname)
    obj.extractall(path=targetdir)
    return targetdir


def clean_msi(folder, pattern, verbose=False, fLOG=print):
    """
    clean all files follwing a specific pattern

    @param      folder      folder
    @param      pattern     files to remove
    @param      verbose     display more information
    @param      fLOG        logging function
    @return                 removed files (as operation)
    """
    def find(loc, pattern):
        exes = []
        for root, dirnames, filenames in os.walk(loc):
            for filename in fnmatch.filter(filenames, pattern):
                exes.append(os.path.join(root, filename))
        return exes
    remove = find(folder, pattern)
    operations = []
    for r in remove:
        if verbose:
            fLOG("[pymy] remove", r)
            os.remove(r)
            operations.append(("remove", r))
    return operations


def extract_copy(fname, targetdir=None, verbose=False, fLOG=print, szip=None):
    """
    Copy *.exe* to targetdir

    @param  fname           local installer (exe)
    @param  targetdir       where to install
    @param  verbose         verbose
    @param  fLOG            logging function
    @param  szip            unused
    @return                 targetdir
    """
    shutil.copy(fname, targetdir)



# -*- coding: utf-8 -*-
from __future__ import print_function
import sys
import os
import warnings
from setuptools import setup, find_packages
from pyquicksetup import read_version, read_readme, default_cmdclass


#########
# settings
#########


project_var_name = "pymyinstall"
project_owner = 'sdpython'
versionPython = "%s.%s" % (sys.version_info.major, sys.version_info.minor)
path = "Lib/site-packages/" + project_var_name
readme = 'README.rst'
history = "HISTORY.rst"
requirements = None

KEYWORDS = [project_var_name, 'installation', 'Xavier Dupré']
DESCRIPTION = """Easy installation of modules for data scientists"""
CLASSIFIERS = [
    'Programming Language :: Python :: 3',
    'Intended Audience :: Developers',
    'Topic :: Scientific/Engineering',
    'Topic :: Education',
    'License :: OSI Approved :: MIT License',
    'Development Status :: 5 - Production/Stable'
]


#######
# data
#######

packages = find_packages('src', exclude='src')
package_dir = {k: "src/" + k.replace(".", "/") for k in packages}
package_data = {
    project_var_name + ".win_installer": ["*.r", "*.jl", "*.iss"],
    project_var_name + ".win_installer.icons": ["*.ico"],
    project_var_name + ".win_installer.tutorial.french": ["*.xlsx"],
}


entry_points = {
    'console_scripts': [
        'pymy_update3 = pymyinstall.cli.pymy_update:main',
        'pymy_install3 = pymyinstall.cli.pymy_install:main',
        'pymy_deps3 = pymyinstall.cli.pymy_deps:main',
        'pymy_status3 = pymyinstall.cli.pymy_status:main',
    ]}
if sys.platform.startswith("win"):
    entry_points['console_scripts'].extend([
        'pymy_update = pymyinstall.cli.pymy_update:main',
        'pymy_install = pymyinstall.cli.pymy_install:main',
        'pymy_deps = pymyinstall.cli.pymy_deps:main',
        'pymy_status = pymyinstall.cli.pymy_status:main',
    ])

setup(
    name=project_var_name,
    version=read_version(__file__, project_var_name, subfolder='src'),
    author='Xavier Dupré',
    author_email='xavier.dupre@gmail.com',
    license="MIT",
    url="http://www.xavierdupre.fr/app/pymyinstall/helpsphinx/index.html",
    download_url="https://github.com/sdpython/pymyinstall/",
    description=DESCRIPTION,
    long_description=read_readme(__file__),
    cmdclass=default_cmdclass(),
    keywords=KEYWORDS,
    classifiers=CLASSIFIERS,
    packages=packages,
    package_dir=package_dir,
    package_data=package_data,
    setup_requires=['pyquicksetup>=0.2'],
    install_requires=["setuptools", "pip>=10.0", "urllib3", "requests"],
    entry_points=entry_points,
)



"""
@brief      test log(time=2s)

skip this test for regular run
"""
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall.packaged import find_module_install


class TestFindModule(unittest.TestCase):

    def test_find_module_install(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        mod = find_module_install("docutils==0.8")
        fLOG(mod)
        assert mod.version == "0.8"


if __name__ == "__main__":
    unittest.main()



"""
@brief      test tree node (time=3s)
"""
import os
import unittest
import warnings
from pyquickhelper.loghelper import fLOG, noLOG
from pyquickhelper.pycode import get_temp_folder
from pymyinstall.installhelper.install_venv_helper import create_virtual_env, run_venv_script
from pymyinstall import ModuleInstall


class TestInstallBuildInstall(unittest.TestCase):

    def test_build_install(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        # module
        ModuleInstall("multi_key_dict", "github", "formiaczek",
                      custom=["build", "install"])

        if __name__ != "__main__":
            warnings.warn("does not work well from a virtual environment")
            return

        temp = get_temp_folder(__file__, "temp_build_install")

        create_virtual_env(temp, fLOG=fLOG)
        setup = os.path.normpath(os.path.join(
            temp, "..", "..", "..", "setup.py"))
        self.assertTrue(os.path.exists(setup))
        folder = os.path.split(setup)[0]
        cmd = "{0} install".format(setup)
        fLOG("CMD: " + cmd)
        cwd = os.getcwd()
        os.chdir(folder)
        try:
            run_venv_script(temp, cmd, fLOG=noLOG, is_cmd=True)
        except Exception as e:
            version = "1.1"
            g = os.path.join(temp, "Lib", "site-packages",
                             "pymyinstall-%s-py3.5.egg" % version, "pymyinstall")
            if not os.path.exists(g):
                raise e
        os.chdir(cwd)

        fLOG("INSTALL: module")
        script = """from pymyinstall import ModuleInstall
                    m = ModuleInstall('multi_key_dict', 'github', 'formiaczek', custom=['build', 'install'])
                    m.download(temp_folder='{0}')
                    m.install(temp_folder='{0}')
                    """.replace("                    ", "").format(temp.replace("\\", "/")).strip(" \n\r")
        fLOG(script)
        run_venv_script(temp, script, fLOG=fLOG)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
@file
@brief Defines different a set of sphinx themes.
"""
import sys
from ..installhelper.module_install import ModuleInstall


def sphinx_theme_set():
    """
    list of sphinx themes, it requires the modules in set *small*
    """
    mod = [
        ModuleInstall(
            'sphinxjp.themes.basicstrap', 'pip', purpose="sphinx theme", usage="SPHINX"),
        ModuleInstall('sphinxjp.themecore', 'pip',
                      purpose="A sphinx theme plugin extension.", usage="SPHINX"),
        ModuleInstall('solar_theme', 'pip',
                      purpose="sphinx theme", usage="SPHINX"),
        ModuleInstall('cloud_sptheme', 'pip',
                      purpose="sphinx theme", usage="SPHINX"),
        ModuleInstall(
            'sphinx_readable_theme', 'pip', purpose="sphinx theme", usage="SPHINX"),
        ModuleInstall(
            "hachibee-sphinx-theme", "pip", mname="hachibee_sphinx_theme", purpose="sphinx theme", usage="SPHINX"),
        ModuleInstall("wild_sphinx_theme", "pip",
                      purpose="sphinx theme", usage="SPHINX"),
        ModuleInstall("sphinx_bootstrap_theme", "pip",
                      purpose="sphinx theme", usage="SPHINX"),
        ModuleInstall(
            "sphinxjp.themes.sphinxjp", "pip", purpose="sphinx theme", usage="SPHINX"),
        ModuleInstall(
            "sphinx_py3doc_enhanced_theme", "pip", purpose="sphinx theme", usage="SPHINX") if sys.version_info[0] >= 3 else None,
        ModuleInstall(
            "sphinx-better-theme", "pip", mname="better", purpose="sphinx theme", usage="SPHINX"),
        ModuleInstall("guzzle_sphinx_theme", "pip",
                      purpose="sphinx theme", usage="SPHINX"),
        ModuleInstall("flyingsphinx", "pip",
                      purpose="sphinx theme", usage="SPHINX/PY2") if sys.version_info[0] == 2 else None,
        ModuleInstall("itcase_sphinx_theme", "pip",
                      purpose="sphinx theme", usage="SPHINX"),
        ModuleInstall("sphinxtrap", "pip",
                      purpose="sphinx theme", usage="SPHINX"),
        ModuleInstall("sphinxcontrib-lunrsearch", "pip", mname="sphinxcontrib.lunrsearch",
                      purpose="simple autocompletion for the sphinx documentation", usage="SPHINX"),
        ModuleInstall("sphinxcontrib-imagesvg", "pip", mname="sphinxcontrib.imagesvg",
                      purpose="support svg in Sphinx", usage="SPHINX"),
        ModuleInstall("zerovm-sphinx-theme", "pip", mname="zerovm_sphinx_theme",
                      purpose="sphinx theme", usage="SPHINX"),
        ModuleInstall("sphinx-corlab-theme", "pip", mname="sphinx_corlab_theme",
                      purpose="sphinx theme", usage="SPHINX"),
        ModuleInstall("sphinxjp.themes.htmlslide", "pip",
                      purpose="sphinx theme", usage="SPHINX"),
        ModuleInstall("sphinxjp.themes.gopher", "pip",
                      purpose="sphinx theme", usage="SPHINX"),
        ModuleInstall("sphinx-gallery", "pip", mname="sphinx_gallery",
                      purpose="gallery for examples", usage="SPHINX"),
        ModuleInstall("docrepr", "pip",
                      purpose="docrepr renders Python docstrings in HTML", usage="SPHINX"),
        ModuleInstall("foundation-sphinx-theme", "pip", mname="foundation_sphinx_theme",
                      purpose="sphinx theme", usage="SPHINX"),
        ModuleInstall("sphinx_theme_pd", "pip",
                      purpose="sphinx theme", usage="SPHINX"),
        ModuleInstall("mdn-sphinx-theme", "pip", mname="mdn_theme",
                      purpose="A Sphinx theme for the Mozilla Developer Network", usage="SPHINX"),
        ModuleInstall("sphinx-redactor-theme", "pip", mname="sphinx_redactor_theme",
                      purpose="sphinx theme", usage="SPHINX"),
        ModuleInstall("sphinx-ioam-theme", "pip", mname="sphinx_ioam_theme",
                      purpose="sphinx theme", usage="SPHINX"),
        ModuleInstall("sphinx-bulma-theme", "pip", mname="sphinx_bulma_theme",
                      purpose="sphinx theme", usage="SPHINX"),
        ModuleInstall("yummy-sphinx-theme", "pip", mname="yummy_sphinx_theme",
                      purpose="sphinx theme", usage="SPHINX"),
        ModuleInstall("sphinx-modern-theme", "pip", mname="sphinx_modern_theme",
                      purpose="sphinx theme", usage="SPHINX"),
        ModuleInstall("sphinx-materialdesign-theme", "pip", mname="sphinx_materialdesign_theme",
                      purpose="sphinx theme", usage="SPHINX"),
        ModuleInstall("sphinx_modern_theme_modified", "pip",
                      purpose="sphinx theme", usage="SPHINX"),
    ]
    return [_ for _ in mod if _ is not None]
